from typing import cast
from unittest.mock import Mock, MagicMock
import weakref

import pytest
from pytest_mock import MockerFixture

from ohmqtt.client import Client
from ohmqtt.connection import Connection, MessageHandlers, InvalidStateError
from ohmqtt.error import MQTTError
from ohmqtt.mqtt_spec import MQTTReasonCode, MQTTQoS
from ohmqtt.packet import (
    MQTTPublishPacket,
    MQTTSubscribePacket,
    MQTTUnsubscribePacket,
    MQTTSubAckPacket,
    MQTTUnsubAckPacket,
    MQTTConnAckPacket,
)
from ohmqtt.property import MQTTPublishProps, MQTTSubscribeProps
from ohmqtt.subscriptions import ReconfiguredError, Subscriptions, RetainPolicy


def dummy_callback(client: Client, packet: MQTTPublishPacket) -> None:
    """Dummy SubscribeCallback for testing."""


@pytest.fixture
def mock_client(mocker: MockerFixture) -> Mock:
    return mocker.Mock(spec=Client)  # type: ignore[no-any-return]


@pytest.fixture
def mock_connection(mocker: MockerFixture) -> Mock:
    conn = mocker.MagicMock(spec=Connection)
    conn.fsm.cond.__enter__.return_value = mocker.Mock()
    conn.fsm.lock.__enter__.return_value = mocker.Mock()
    return conn  # type: ignore[no-any-return]


@pytest.fixture
def mock_handlers(mocker: MockerFixture) -> MagicMock:
    return mocker.MagicMock(spec=MessageHandlers)  # type: ignore[no-any-return]


def test_subscriptions_registration(mock_client: Mock, mock_connection: Mock) -> None:
    handlers = MessageHandlers()
    with handlers:
        subscriptions = Subscriptions(handlers, mock_connection, weakref.ref(mock_client))
    assert subscriptions.handle_suback in handlers.get_handlers(MQTTSubAckPacket)
    assert subscriptions.handle_unsuback in handlers.get_handlers(MQTTUnsubAckPacket)


@pytest.mark.parametrize("max_qos", [MQTTQoS.Q0, MQTTQoS.Q1, MQTTQoS.Q2])
@pytest.mark.parametrize("no_local", [True, False])
@pytest.mark.parametrize("retain_as_published", [True, False])
@pytest.mark.parametrize("retain_policy", [RetainPolicy.NEVER, RetainPolicy.ONCE, RetainPolicy.ALWAYS])
def test_subscriptions_subscribe_opts(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock,
    max_qos: MQTTQoS,
    no_local: bool,
    retain_as_published: bool,
    retain_policy: RetainPolicy
) -> None:
    """Test subscribing with options."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    subscriptions.subscribe(
        "test/topic",
        dummy_callback,
        max_qos=max_qos,
        share_name="test_share",
        no_local=no_local,
        retain_as_published=retain_as_published,
        retain_policy=retain_policy,
        sub_id=23,
        user_properties=[("key", "value")],
    )
    expected_opts = max_qos.value | (retain_policy << 4) | (retain_as_published << 3) | (no_local << 2)
    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        topics=[("$share/test_share/test/topic", expected_opts)],
        packet_id=1,
        properties=MQTTSubscribeProps(
            SubscriptionIdentifier={23},
            UserProperty=(("key", "value"),),
        ),
    ))


def test_subscriptions_multi_subscribe(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    """Test subscribing multiple times to the same topic before acknowledgement."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    handle1 = subscriptions.subscribe("test/topic", dummy_callback)
    assert handle1 is not None

    handle2 = subscriptions.subscribe("test/topic", dummy_callback)
    assert handle2 is not None
    assert handle1 != handle2

    with pytest.raises(ReconfiguredError):
        handle1.wait_for_ack(timeout=0.001)
    assert handle1.ack is None

    # Simulate receiving a SUBACK packet
    suback_packet = MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    )
    subscriptions.handle_suback(suback_packet)

    assert handle2.wait_for_ack(timeout=0.1) == suback_packet
    assert handle2.ack == suback_packet


def test_subscriptions_resubscribe(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock,
) -> None:
    """Test reconfiguring an existing and acknowledged subscription."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    subscriptions.subscribe("test/topic", dummy_callback, no_local=True)
    expected_opts = 2 | (True << 2)
    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        topics=[("test/topic", expected_opts)],
        packet_id=1,
    ))
    mock_connection.reset_mock()
    subscriptions.handle_suback(MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    ))

    subscriptions.subscribe("test/topic", dummy_callback, no_local=False)
    expected_opts = 2
    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        topics=[("test/topic", expected_opts)],
        packet_id=2,
    ))
    mock_connection.reset_mock()
    subscriptions.handle_suback(MQTTSubAckPacket(
        packet_id=2,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    ))


def test_subscriptions_fast_subscribe(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    """Test subscribing and unsubscribing quickly."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    subscriptions.subscribe("test/topic", dummy_callback)
    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        topics=[("test/topic", 2)],
        packet_id=1,
    ))
    mock_connection.reset_mock()

    subscriptions.unsubscribe("test/topic")
    mock_connection.send.assert_called_once_with(MQTTUnsubscribePacket(
        topics=["test/topic"],
        packet_id=1,
    ))
    mock_connection.reset_mock()

    subscriptions.handle_suback(MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    ))

    # We should not send another UNSUBSCRIBE packet after the SUBACK
    subscriptions.unsubscribe("test/topic")
    mock_connection.send.assert_not_called()

    subscriptions.subscribe("test/topic", dummy_callback)
    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        topics=[("test/topic", 2)],
        packet_id=2,
    ))
    mock_connection.reset_mock()

    subscriptions.handle_unsuback(MQTTUnsubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.Success],
    ))

    subscriptions.handle_suback(MQTTSubAckPacket(
        packet_id=2,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    ))


def test_subscriptions_wait_for_suback(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    """Test waiting for a SUBACK packet."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    handle = subscriptions.subscribe("test/topic", dummy_callback)
    assert handle is not None
    with pytest.raises(TimeoutError):
        handle.wait_for_ack(timeout=0.001)

    # Simulate receiving a SUBACK packet
    suback_packet = MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    )
    subscriptions.handle_suback(suback_packet)

    assert handle.wait_for_ack(timeout=0.001) == suback_packet
    assert handle.ack == suback_packet


@pytest.mark.parametrize("ack_rc", [MQTTReasonCode.Success, MQTTReasonCode.UnspecifiedError])
def test_subscriptions_wait_for_unsuback(
    ack_rc: MQTTReasonCode,
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    """Test waiting for a SUBACK packet."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    subscriptions.subscribe("test/topic", dummy_callback)

    # Simulate receiving a SUBACK packet
    suback_packet = MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[ack_rc],
    )
    subscriptions.handle_suback(suback_packet)

    handle = subscriptions.unsubscribe("test/topic")
    assert handle is not None
    with pytest.raises(TimeoutError):
        handle.wait_for_ack(timeout=0.001)

    # Simulate receiving an UNSUBACK packet
    unsuback_packet = MQTTUnsubAckPacket(
        packet_id=1,
        reason_codes=[ack_rc],
    )
    subscriptions.handle_unsuback(unsuback_packet)

    assert handle.wait_for_ack(timeout=0.001) == unsuback_packet
    assert handle.ack == unsuback_packet


@pytest.mark.parametrize("session_present", [True, False])
def test_subscriptions_subscribe_failure(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock,
    session_present: bool
) -> None:
    """Test replaying SUBSCRIBE packets on reconnection after calling in a bad state."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    mock_connection.send.side_effect = InvalidStateError("TEST")
    handle = subscriptions.subscribe("test/topic", dummy_callback)

    assert handle is not None

    mock_connection.send.side_effect = None
    mock_connection.reset_mock()

    subscriptions.handle_connack(MQTTConnAckPacket(session_present=session_present))

    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        topics=[("test/topic", 2)],
        packet_id=1,
    ))
    mock_connection.reset_mock()

    subscriptions.handle_connack(MQTTConnAckPacket(session_present=session_present))

    if not session_present:
        # If the session was not present, we should send the SUBSCRIBE packet again.
        mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
            topics=[("test/topic", 2)],
            packet_id=1,
        ))


def test_subscriptions_unsubscribe_failure(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    subscriptions.subscribe("test/topic", dummy_callback)
    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        topics=[("test/topic", 2)],
        packet_id=1,
    ))
    subscriptions.handle_suback(MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    ))

    mock_connection.send.side_effect = InvalidStateError("TEST")
    handle = subscriptions.unsubscribe("test/topic")

    assert handle is not None

    mock_connection.send.side_effect = None
    mock_connection.reset_mock()

    subscriptions.handle_connack(MQTTConnAckPacket(session_present=True))

    mock_connection.send.assert_called_once_with(MQTTUnsubscribePacket(
        topics=["test/topic"],
        packet_id=1,
    ))
    mock_connection.reset_mock()


def test_subscriptions_unsubscribe_null(mock_handlers: MagicMock, mock_client: Mock, mock_connection: Mock) -> None:
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    handle = subscriptions.unsubscribe("test/topic")
    assert handle is None


def test_subscriptions_unsubscribe_redundant(mock_handlers: MagicMock, mock_client: Mock, mock_connection: Mock) -> None:
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    subscriptions.subscribe("test/topic", dummy_callback)
    subscriptions.handle_suback(MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    ))

    handle1 = subscriptions.unsubscribe("test/topic")
    handle2 = subscriptions.unsubscribe("test/topic")
    assert handle1 is not None
    subscriptions.handle_unsuback(MQTTUnsubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    ))
    assert handle2 is not None
    assert handle2.ack is not None

    mock_connection.reset_mock()
    assert subscriptions.unsubscribe("test/topic") is None
    mock_connection.send.assert_not_called()


@pytest.mark.parametrize(("tf", "topic"), [
    ("test/topic", "test/topic"),
    ("test/+", "test/topic"),
    ("#", "test/topic"),
])
def test_subscriptions_handle_publish(
    tf: str,
    topic: str,
    mock_handlers: MagicMock,
    mock_connection: Mock
) -> None:
    """Test handling a publish packet."""
    mock_client = cast(Client, frozenset())
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    recvd = []
    def callback(client: Client, packet: MQTTPublishPacket) -> None:
        recvd.append(packet)
    subscriptions.subscribe(tf, callback)

    publish_packet = MQTTPublishPacket(
        topic=topic,
        payload=b"test payload",
    )
    subscriptions.handle_publish(publish_packet)

    assert recvd == [publish_packet]

    del mock_client
    with pytest.raises(RuntimeError, match="Client went out of scope"):
        subscriptions.handle_publish(publish_packet)


def test_subscriptions_handle_publish_shared(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    """Test handling a publish packet with shared subscriptions."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    recvd1 = []
    recvd2 = []
    def callback1(client: Client, packet: MQTTPublishPacket) -> None:
        recvd1.append(packet)
    def callback2(client: Client, packet: MQTTPublishPacket) -> None:
        recvd2.append(packet)

    subscriptions.subscribe("test/topic", callback1)
    subscriptions.subscribe("test/topic", callback2, share_name="test_share")

    publish_packet = MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
    )
    subscriptions.handle_publish(publish_packet)

    assert recvd1 == [publish_packet]
    assert recvd2 == [publish_packet]


def test_subscriptions_handle_publish_sub_id(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    """Test handling a publish packet with subscription identifier."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    recvd1 = []
    recvd2 = []
    def callback1(client: Client, packet: MQTTPublishPacket) -> None:
        recvd1.append(packet)
    def callback2(client: Client, packet: MQTTPublishPacket) -> None:
        recvd2.append(packet)

    subscriptions.subscribe("test/topic", callback1)
    subscriptions.subscribe("test/+", callback2, sub_id=1)

    publish_packet = MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
        properties=MQTTPublishProps(SubscriptionIdentifier={1}),
    )
    subscriptions.handle_publish(publish_packet)

    assert recvd1 == []
    assert recvd2 == [publish_packet]


def test_subscriptions_handle_publish_exception(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    """Test handling a publish packet with a broken callback."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    recvd = []
    def callback1(client: Client, packet: MQTTPublishPacket) -> None:
        raise RuntimeError("TEST")
    def callback2(client: Client, packet: MQTTPublishPacket) -> None:
        recvd.append(packet)

    subscriptions.subscribe("test/topic", callback1)
    subscriptions.subscribe("test/+", callback2)

    publish_packet = MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
    )
    subscriptions.handle_publish(publish_packet)

    assert recvd == [publish_packet]


def test_subscriptions_packet_id(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    # Test the private methods because the public interface is slow.
    for n in range(0xffff + 5):
        assert subscriptions._get_next_sub_packet_id() == (n % 0xffff) + 1  # noqa: SLF001

    for n in range(0xffff + 5):
        assert subscriptions._get_next_unsub_packet_id() == (n % 0xffff) + 1  # noqa: SLF001


def test_subscriptions_ack_unknown(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    with pytest.raises(MQTTError, match="Received SUBACK for unknown packet ID: 5") as excinfo:
        subscriptions.handle_suback(MQTTSubAckPacket(
            packet_id=5,
            reason_codes=[MQTTReasonCode.GrantedQoS2],
        ))
    assert excinfo.value.reason_code == MQTTReasonCode.ProtocolError
    with pytest.raises(MQTTError, match="Received UNSUBACK for unknown packet ID: 6") as excinfo:
        subscriptions.handle_unsuback(MQTTUnsubAckPacket(
            packet_id=6,
            reason_codes=[MQTTReasonCode.Success],
        ))
    assert excinfo.value.reason_code == MQTTReasonCode.ProtocolError


def test_subscriptions_handle_session_reset(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock,
) -> None:
    """Test resetting the session."""
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))

    subscriptions.subscribe("test/topic1", dummy_callback)
    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        packet_id=1,
        topics=[("test/topic1", 2)],
    ))
    mock_connection.send.reset_mock()

    subscriptions.handle_connack(MQTTConnAckPacket(
        session_present=True,
        reason_code=MQTTReasonCode.Success,
    ))

    # Should replay the unacked subscribe on reconnection.
    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        packet_id=1,
        topics=[("test/topic1", 2)],
    ))
    mock_connection.send.reset_mock()

    subscriptions.handle_suback(MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    ))

    subscriptions.handle_connack(MQTTConnAckPacket(
        session_present=True,
        reason_code=MQTTReasonCode.Success,
    ))

    # Should not replay the acked subscribe on reconnection with session present.
    mock_connection.send.assert_not_called()

    subscriptions.handle_connack(MQTTConnAckPacket(
        session_present=False,
        reason_code=MQTTReasonCode.Success,
    ))

    # Should replay the acked subscribe on reconnection with no session present.
    mock_connection.send.assert_called_once_with(MQTTSubscribePacket(
        packet_id=1,
        topics=[("test/topic1", 2)],
    ))
    mock_connection.send.reset_mock()

    subscriptions.handle_suback(MQTTSubAckPacket(
        packet_id=1,
        reason_codes=[MQTTReasonCode.GrantedQoS2],
    ))

    subscriptions.unsubscribe("test/topic1")
    mock_connection.send.assert_called_once_with(MQTTUnsubscribePacket(
        packet_id=1,
        topics=["test/topic1"],
    ))
    mock_connection.send.reset_mock()

    subscriptions.handle_connack(MQTTConnAckPacket(
        session_present=True,
        reason_code=MQTTReasonCode.Success,
    ))

    # Replay unsubscribe if not acked before connection reset.
    mock_connection.send.assert_called_once_with(MQTTUnsubscribePacket(
        packet_id=1,
        topics=["test/topic1"],
    ))
    mock_connection.send.reset_mock()

    subscriptions.handle_connack(MQTTConnAckPacket(
        session_present=False,
        reason_code=MQTTReasonCode.Success,
    ))

    # Don't bother replaying unsubscribe after session has been reset.
    mock_connection.send.assert_not_called()


def test_subscriptions_slots(
    mock_handlers: MagicMock,
    mock_client: Mock,
    mock_connection: Mock
) -> None:
    subscriptions = Subscriptions(mock_handlers, mock_connection, weakref.ref(mock_client))
    # Slots all the way down.
    assert not hasattr(subscriptions, "__dict__")
    assert hasattr(subscriptions, "__weakref__")
    assert all(hasattr(subscriptions, attr) for attr in subscriptions.__slots__)



from .client import Client as Client
from .error import MQTTError as MQTTError
from .handles import PublishHandle as PublishHandle
from .handles import SubscribeHandle as SubscribeHandle
from .handles import UnsubscribeHandle as UnsubscribeHandle
from .logger import set_log_level as set_log_level
from .mqtt_spec import MQTTReasonCode as MQTTReasonCode
from .mqtt_spec import MQTTQoS as MQTTQoS
from .packet import MQTTPacket as MQTTPacket
from .packet import MQTTAuthPacket as MQTTAuthPacket
from .packet import MQTTConnectPacket as MQTTConnectPacket
from .packet import MQTTConnAckPacket as MQTTConnAckPacket
from .packet import MQTTDisconnectPacket as MQTTDisconnectPacket
from .packet import MQTTPublishPacket as MQTTPublishPacket
from .packet import MQTTPubAckPacket as MQTTPubAckPacket
from .packet import MQTTPubRecPacket as MQTTPubRecPacket
from .packet import MQTTPubRelPacket as MQTTPubRelPacket
from .packet import MQTTPubCompPacket as MQTTPubCompPacket
from .packet import MQTTSubscribePacket as MQTTSubscribePacket
from .packet import MQTTSubAckPacket as MQTTSubAckPacket
from .packet import MQTTUnsubscribePacket as MQTTUnsubscribePacket
from .packet import MQTTUnsubAckPacket as MQTTUnsubAckPacket
from .packet import MQTTPingReqPacket as MQTTPingReqPacket
from .packet import MQTTPingRespPacket as MQTTPingRespPacket
from .persistence.base import LostMessageError as LostMessageError
from .property import MQTTProperties as MQTTProperties
from .property import MQTTAuthProps as MQTTAuthProps
from .property import MQTTConnectProps as MQTTConnectProps
from .property import MQTTConnAckProps as MQTTConnAckProps
from .property import MQTTDisconnectProps as MQTTDisconnectProps
from .property import MQTTPublishProps as MQTTPublishProps
from .property import MQTTPubAckProps as MQTTPubAckProps
from .property import MQTTPubRecProps as MQTTPubRecProps
from .property import MQTTPubRelProps as MQTTPubRelProps
from .property import MQTTPubCompProps as MQTTPubCompProps
from .property import MQTTSubscribeProps as MQTTSubscribeProps
from .property import MQTTSubAckProps as MQTTSubAckProps
from .property import MQTTUnsubscribeProps as MQTTUnsubscribeProps
from .property import MQTTUnsubAckProps as MQTTUnsubAckProps
from .property import MQTTWillProps as MQTTWillProps
from .subscriptions import RetainPolicy as RetainPolicy
from .topic_alias import AliasPolicy as AliasPolicy
from .topic_alias import MaxOutboundAliasError as MaxOutboundAliasError

from importlib.metadata import version

__version__ = version("ohmqtt")
__version_info__ = tuple(int(x) for x in __version__.split(".") if x.isdigit())



from __future__ import annotations

import ssl
import sys
import threading
from typing import Final, Sequence
import weakref

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from .connection import Address, ConnectParams, Connection, MessageHandlers
from .handles import PublishHandle, SubscribeHandle, UnsubscribeHandle
from .logger import get_logger
from .mqtt_spec import MQTTReasonCode, MQTTQoS
from .packet import MQTTAuthPacket
from .property import MQTTAuthProps, MQTTConnectProps, MQTTPublishProps, MQTTWillProps
from .session import Session
from .subscriptions import Subscriptions, SubscribeCallback, RetainPolicy
from .topic_alias import AliasPolicy

logger: Final = get_logger("client")


class Client:
    """High level interface for the MQTT client.

    :param db_path: Path to the database file for persistence.  If none provided, in-memory store will be used.
    :param db_fast: If True, use a faster database implementation (e.g. SQLite WAL).
    """
    __slots__ = (
        "__weakref__",
        "_thread",
        "connection",
        "session",
    )

    def __init__(self, db_path: str = "", *, db_fast: bool = False) -> None:
        self._thread: threading.Thread | None = None
        message_handlers = MessageHandlers()
        with message_handlers as handlers:
            self.connection = Connection(handlers)
            subscriptions = Subscriptions(handlers, self.connection, weakref.ref(self))
            self.session = Session(handlers, subscriptions, self.connection, db_path=db_path, db_fast=db_fast)
            handlers.register(MQTTAuthPacket, self.handle_auth)

    def __enter__(self) -> Self:
        self.start_loop()
        return self

    def __exit__(self, *args: object) -> None:
        self.shutdown()

    def connect(
        self,
        address: str,
        *,
        client_id: str = "",
        clean_start: bool = False,
        connect_timeout: float | None = None,
        reconnect_delay: int = 0,
        keepalive_interval: int = 0,
        tcp_nodelay: bool = True,
        tls_context: ssl.SSLContext | None = None,
        tls_hostname: str = "",
        username: str | None = None,
        password: bytes | None = None,
        will_topic: str = "",
        will_payload: bytes = b"",
        will_qos: int = 0,
        will_retain: bool = False,
        will_properties: MQTTWillProps | None = None,
        connect_properties: MQTTConnectProps | None = None,
    ) -> None:
        """Connect to the broker.

        :param address: The address of the broker to connect to.
            Addresses may be in the form:
            - mqtt://host:port      (TCP)
            - mqtts://host:port     (TLS)
            - ws://host:port        (WebSocket)
            - wss://host:port       (WebSocket/TLS)
            - unix:/path/to/socket  (if supported on the platform)
            - When the protocol is omitted, mqtt:// is assumed.
            - When the port is omitted, the default port for the protocol is used.
                - mqtt: 1883
                - mqtts: 8883
                - ws: 80
                - wss: 443
        :param client_id: The client ID to use for the connection, or empty string to request one from the broker.
        :param clean_start: If True, an existing session will not be resumed.
        :param connect_timeout: Timeout for the connection attempt in seconds, or None for no timeout.
        :param reconnect_delay: Delay in seconds before attempting to reconnect after a disconnection, or 0 to disable reconnect.
        :param keepalive_interval: The keep alive interval in seconds, or 0 to disable keep alive.
        :param tcp_nodelay: If True, enable TCP_NODELAY to disable Nagle's algorithm.
        :param tls_context: An SSLContext for TLS connections, or None to use the default.
        :param tls_hostname: The hostname to use for TLS connections, or empty string to determine from the address.
        :param username: The username for MQTT authentication, or None to disable.
        :param password: The password for MQTT authentication, or None to disable.
        :param will_topic: The topic for the Will message, or empty string to disable.
        :param will_payload: The payload for the Will message.
        :param will_qos: The QoS level for the Will message (0, 1, or 2).
        :param will_retain: If True, the Will message will be retained.
        :param will_properties: Properties for the Will message.
        :param connect_properties: Properties for the CONNECT packet.
        """
        _address = Address(address)
        params = ConnectParams(
            address=_address,
            client_id=client_id,
            clean_start=clean_start,
            connect_timeout=connect_timeout,
            reconnect_delay=reconnect_delay,
            keepalive_interval=keepalive_interval,
            tcp_nodelay=tcp_nodelay,
            tls_context=tls_context,
            tls_hostname=tls_hostname,
            username=username,
            password=password,
            will_topic=will_topic,
            will_payload=will_payload,
            will_qos=will_qos,
            will_retain=will_retain,
            will_properties=will_properties if will_properties is not None else MQTTWillProps(),
            connect_properties=connect_properties if connect_properties is not None else MQTTConnectProps(),
        )
        with self.connection.fsm.lock:
            self.connection.connect(params)
            self.session.set_params(params)

    def disconnect(self) -> None:
        """Disconnect from the broker."""
        self.connection.disconnect()

    def shutdown(self) -> None:
        """Shutdown the client and close the connection."""
        self.connection.shutdown()

    def publish(
        self,
        topic: str,
        payload: bytes | bytearray | str,
        *,
        qos: int | MQTTQoS = MQTTQoS.Q0,
        retain: bool = False,
        properties: MQTTPublishProps | None = None,
        alias_policy: AliasPolicy = AliasPolicy.NEVER,
    ) -> PublishHandle:
        """Publish a message to a topic.

        :param topic: The topic to publish to.
        :param payload: The payload of the message. If a string is provided, it will be encoded as UTF-8.
        :param qos: The QoS level for the message (0, 1, or 2).
        :param retain: If True, the message will be retained by the broker.
        :param properties: Properties for the PUBLISH packet.
        :param alias_policy: The policy for using automatic topic aliases.
        """
        if isinstance(payload, str):
            payload = payload.encode("utf-8")
        elif not isinstance(payload, bytes):
            payload = bytes(payload)
        if not isinstance(qos, MQTTQoS):
            qos = MQTTQoS(qos)
        properties = properties if properties is not None else None
        return self.session.publish(
            topic,
            payload,
            qos=qos,
            retain=retain,
            properties=properties,
            alias_policy=alias_policy,
        )

    def subscribe(
        self,
        topic_filter: str,
        callback: SubscribeCallback,
        max_qos: int | MQTTQoS = MQTTQoS.Q2,
        *,
        share_name: str | None = None,
        no_local: bool = False,
        retain_as_published: bool = False,
        retain_policy: RetainPolicy = RetainPolicy.ALWAYS,
        sub_id: int | None = None,
        user_properties: Sequence[tuple[str, str]] | None = None,
    ) -> SubscribeHandle | None:
        """Subscribe to a topic filter with a callback.

        If the client is connected, returns a handle which can be used to unsubscribe from the topic filter
        or wait for the subscription to be acknowledged.

        If the client is not connected, returns None.

        :param topic_filter: The topic filter to subscribe to.
        :param callback: The callback to call when a message is received on the subscribed topic.
        :param max_qos: The maximum QoS level for the subscription (0, 1, or 2).
        :param share_name: The name of a shared subscription to use.
        :param no_local: If True, do not receive messages published by this client.
        :param retain_as_published: If True, the retain flag of messages will match the original message.
        :param retain_policy: The policy for retained messages.
        :param sub_id: An optional subscription ID for the subscription.
        :param user_properties: Optional user properties to include in the subscription request.
        """
        if not isinstance(max_qos, MQTTQoS):
            max_qos = MQTTQoS(max_qos)
        return self.session.subscriptions.subscribe(
            topic_filter,
            callback,
            max_qos=max_qos,
            share_name=share_name,
            no_local=no_local,
            retain_as_published=retain_as_published,
            retain_policy=retain_policy,
            sub_id=sub_id,
            user_properties=user_properties,
        )

    def unsubscribe(
        self,
        topic_filter: str,
        *,
        share_name: str | None = None,
    ) -> UnsubscribeHandle | None:
        """Unsubscribe from a topic filter.

        If the client is connected, returns a handle which can be used to wait for the unsubscription to be acknowledged.

        If the client is not connected, returns None.

        :param topic_filter: The topic filter to unsubscribe from.
        :param share_name: The name of a shared subscription to use.
        """
        return self.session.subscriptions.unsubscribe(
            topic_filter,
            share_name=share_name,
        )

    def auth(
        self,
        *,
        authentication_method: str | None = None,
        authentication_data: bytes | None = None,
        reason_string: str | None = None,
        user_properties: Sequence[tuple[str, str]] | None = None,
        reason_code: MQTTReasonCode = MQTTReasonCode.Success,
    ) -> None:
        """Send an AUTH packet to the broker.

        :param authentication_method: The authentication method to use.
        :param authentication_data: Authentication data to send.
        :param reason_string: A reason string to include in the AUTH packet.
        :param user_properties: Optional user properties to include in the AUTH packet.
        :param reason_code: The reason code for the AUTH packet.
        """
        properties = MQTTAuthProps()
        if authentication_method is not None:
            properties.AuthenticationMethod = authentication_method
        if authentication_data is not None:
            properties.AuthenticationData = authentication_data
        if reason_string is not None:
            properties.ReasonString = reason_string
        if user_properties is not None:
            properties.UserProperty = user_properties
        packet = MQTTAuthPacket(
            reason_code=reason_code,
            properties=properties,
        )
        self.connection.send(packet)

    def wait_for_connect(self, timeout: float | None = None) -> None:
        """Wait for the client to connect to the broker.

        :raises TimeoutError: The timeout was exceeded."""
        if not self.connection.wait_for_connect(timeout):
            raise TimeoutError("Waiting for connection timed out")

    def wait_for_disconnect(self, timeout: float | None = None) -> None:
        """Wait for the client to disconnect from the broker.

        :raises TimeoutError: The timeout was exceeded."""
        if not self.connection.wait_for_disconnect(timeout):
            raise TimeoutError("Waiting for disconnection timed out")

    def wait_for_shutdown(self, timeout: float | None = None) -> None:
        """Wait for the client to disconnect and finalize.

        :raises TimeoutError: The timeout was exceeded."""
        if not self.connection.wait_for_shutdown(timeout):
            raise TimeoutError("Waiting for disconnection timed out")

    def start_loop(self) -> None:
        """Start the client state machine in a separate thread."""
        if self._thread is not None:
            raise RuntimeError("Connection loop already started")
        self._thread = threading.Thread(target=self.loop_forever, daemon=True)
        self._thread.start()

    def loop_once(self, max_wait: float | None = 0.0) -> None:
        """Run a single iteration of the MQTT client loop.

        If max_wait is 0.0 (the default), this call will not block.

        If max_wait is None, this call will block until the next event.

        Any other numeric max_wait value may block for maximum that amount of time in seconds."""
        self.connection.loop_once(max_wait)

    def loop_forever(self) -> None:
        """Run the MQTT client loop.

        This will run until the client is stopped or shutdown.
        """
        self.connection.loop_forever()

    def loop_until_connected(self, timeout: float | None = None) -> bool:
        """Run the MQTT client loop until the client is connected to the broker.

        If a timeout is provided, the loop will give up after that amount of time.

        Returns True if the client is connected, False if the timeout was reached."""
        return self.connection.loop_until_connected(timeout)

    def is_connected(self) -> bool:
        """Check if the client is connected to the broker."""
        return self.connection.is_connected()

    def handle_auth(self, packet: MQTTAuthPacket) -> None:
        """Callback for an AUTH packet from the broker."""
        logger.debug("Got an AUTH packet")



import argparse


parser = argparse.ArgumentParser(description="ohmqtt example")
parser.add_argument(
    "-a", "--address",
    type=str,
    default="localhost:1883",
    help="MQTT broker address (default: localhost:1883)",
)



from __future__ import annotations

import ssl
import socket
from typing import Final, TYPE_CHECKING

from .base import FSMState
from .closed import ClosedState
from ..types import ConnectParams, StateData, StateEnvironment
from ...logger import get_logger
from ...mqtt_spec import MQTTReasonCode

if TYPE_CHECKING:
    from ..fsm import FSM

logger: Final = get_logger("ohmqtt.connection.states.closing")


class ClosingState(FSMState):
    """Gracefully closing the connection.

    The socket will be shutdown for reading, but existing buffers will be flushed."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        # Avoid circular imports.
        from .connected import ConnectedState  # noqa: PLC0415

        if fsm.previous_state == ConnectedState:
            logger.debug("Gracefully closing connection")
            if state_data.disconnect_rc is None:
                state_data.disconnect_rc = MQTTReasonCode.NormalDisconnection
        else:
            logger.debug("Skipping ClosingState")
            fsm.change_state(ClosedState)
            return

        state_data.timeout.interval = params.connect_timeout
        state_data.timeout.mark()

        # Shutdown only the read side of the socket, so we can still send data.
        # We don't want to shutdown the socket if we're using TLS, it would cause the protocol to fail.
        if not params.address.use_tls:
            try:
                state_data.sock.shutdown(socket.SHUT_RD)
            except OSError as exc:
                logger.debug("Error while shutting down socket reading: %s", exc)

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        # Wait for the socket to be writable.
        if state_data.timeout.exceeded():
            logger.error("ClosingState timed out")
            fsm.change_state(ClosedState)
            return True

        if not state_data.write_buffer:
            logger.debug("No more data to send, connection closed")
            fsm.change_state(ClosedState)
            return True

        with fsm.selector:
            timeout = state_data.timeout.get_timeout(max_wait)
            _, writable = fsm.selector.select(write=True, timeout=timeout)

        if writable:
            try:
                sent = state_data.sock.send(state_data.write_buffer)
                del state_data.write_buffer[:sent]
                state_data.keepalive.mark_send()
            except (BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError):
                pass
            except BrokenPipeError:
                logger.error("Socket lost while closing")
                fsm.change_state(ClosedState)
                return True
        return False



"""PINGREQ and PINGRESP packets."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Final

from .base import MQTTPacket
from ..error import MQTTError
from ..mqtt_spec import MQTTPacketType, MQTTReasonCode


@dataclass(match_args=True, slots=True)
class MQTTPingReqPacket(MQTTPacket):
    packet_type = MQTTPacketType.PINGREQ

    def __str__(self) -> str:
        return "PINGREQ[]"

    def encode(self) -> bytes:
        return b"\xc0\x00"  # PINGREQ is a fixed header with no payload.

    @classmethod
    def decode(cls, flags: int, data: memoryview) -> MQTTPingReqPacket:
        if flags != 0:
            raise MQTTError(f"Invalid flags, expected 0 but got {flags}", MQTTReasonCode.MalformedPacket)
        if len(data) != 0:
            raise MQTTError(f"Invalid length, expected 0 but got {len(data)}", MQTTReasonCode.MalformedPacket)
        return MQTTPingReqPacket()


@dataclass(match_args=True, slots=True)
class MQTTPingRespPacket(MQTTPacket):
    packet_type = MQTTPacketType.PINGRESP

    def __str__(self) -> str:
        return "PINGRESP[]"

    def encode(self) -> bytes:
        return b"\xd0\x00"  # PINGRESP is a fixed header with no payload.

    @classmethod
    def decode(cls, flags: int, data: memoryview) -> MQTTPingRespPacket:
        if flags != 0:
            raise MQTTError(f"Invalid flags, expected 0 but got {flags}", MQTTReasonCode.MalformedPacket)
        if len(data) != 0:
            raise MQTTError(f"Invalid length, expected 0 but got {len(data)}", MQTTReasonCode.MalformedPacket)
        return MQTTPingRespPacket()


# Pre-encoded packets for PINGREQ and PINGRESP
PING: Final = MQTTPingReqPacket().encode()
PONG: Final = MQTTPingRespPacket().encode()



from __future__ import annotations

import selectors
import socket
import ssl
import threading
from typing import Final, TypeAlias

from ..logger import get_logger
from ..protected import Protected, protect, LockLike

logger: Final = get_logger("connection.selector")

SocketLike: TypeAlias = socket.socket | ssl.SSLSocket


class InterruptibleSelector(Protected):
    """A select() method which can be interrupted by a call to interrupt().

    This can be used to interrupt a blocking select() call from another thread."""
    __slots__ = (
        "_closed",
        "_in_select",
        "_interrupt_r",
        "_interrupt_w",
        "_interrupted",
        "_selector",
        "_sock",
    )

    def __init__(self, lock: LockLike | None = None) -> None:
        super().__init__(lock if lock is not None else threading.RLock())
        self._selector = selectors.DefaultSelector()
        self._sock: SocketLike | None = None
        self._closed = False
        self._in_select = False
        self._interrupted = False
        self._interrupt_r, self._interrupt_w = socket.socketpair()
        self._interrupt_r.setblocking(False)
        self._interrupt_w.setblocking(False)
        self._selector.register(self._interrupt_r, selectors.EVENT_READ)

    def _drain(self) -> None:
        """Drain the interrupt socket."""
        data = self._interrupt_r.recv(2)
        assert len(data) <= 1, "Drained more than one interrupt"

    @protect
    def close(self) -> None:
        """Finalize this instance."""
        self._closed = True
        self._interrupt_r.close()
        self._interrupt_w.close()
        self._selector.close()

    @protect
    def interrupt(self) -> None:
        """Interrupt the select call, if we are in select."""
        if self._closed:
            raise RuntimeError("Selector is closed")
        if self._in_select and not self._interrupted:
            # We are in select, send an interrupt.
            self._interrupted = True
            self._interrupt_w.send(b"\x00")

    def change_sock(self, sock: SocketLike) -> None:
        """Change the socket for the selector."""
        if self._closed:
            raise RuntimeError("Selector is closed")
        if sock == self._sock:
            return
        if self._sock is not None:
            self._selector.unregister(self._sock)
        self._selector.register(sock, selectors.EVENT_READ)
        self._sock = sock

    @protect
    def select(
        self,
        *,
        read: bool = False,
        write: bool = False,
        timeout: float | None = None,
    ) -> tuple[bool, bool]:
        """Select an optional socket with a timeout, allowing for interruption.

        This method must be called with the lock already held."""
        if self._closed:
            raise RuntimeError("InterruptibleSelector is closed")
        if self._sock is None:
            raise RuntimeError("No socket set for InterruptibleSelector")
        wanted_events = (read * selectors.EVENT_READ) | (write * selectors.EVENT_WRITE)
        if wanted_events == 0:
            raise ValueError("Must select either read or write or both")
        self._selector.modify(self._sock, wanted_events)
        readable, writable = False, False
        self._in_select = True
        self._interrupted = False
        self.release()
        try:
            events = self._selector.select(timeout)
            for key, event in events:
                if key.fileobj is self._sock:
                    readable = bool(event & selectors.EVENT_READ)
                    writable = bool(event & selectors.EVENT_WRITE)
        finally:
            self.acquire()
            if self._interrupted:
                self._drain()
            self._in_select = False
            self._interrupted = False
        return readable, writable



"""CONNECT, CONNACK, and DISCONNECT packets."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Final

from .base import MQTTPacket
from ..error import MQTTError
from ..mqtt_spec import MQTTReasonCode, MQTTPacketType
from ..property import (
    MQTTConnectProps,
    MQTTConnAckProps,
    MQTTDisconnectProps,
    MQTTWillProps,
)
from ..serialization import (
    encode_string,
    encode_binary,
    encode_uint8,
    encode_uint16,
    encode_bool,
    encode_varint,
    decode_string,
    decode_binary,
    decode_uint8,
    decode_uint16,
    decode_bool,
)


HEAD_CONNECT: Final = MQTTPacketType.CONNECT << 4
HEAD_CONNACK: Final = MQTTPacketType.CONNACK << 4
HEAD_DISCONNECT: Final = MQTTPacketType.DISCONNECT << 4


@dataclass(match_args=True, slots=True)
class MQTTConnectPacket(MQTTPacket):
    packet_type = MQTTPacketType.CONNECT
    props_type = MQTTConnectProps
    client_id: str = ""
    keep_alive: int = 0
    protocol_version: int = 5
    clean_start: bool = False
    will_props: MQTTWillProps = field(default_factory=MQTTWillProps)
    will_topic: str = ""
    will_payload: bytes = b""
    will_qos: int = 0
    will_retain: bool = False
    username: str | None = None
    password: bytes | None = None
    properties: MQTTConnectProps = field(default_factory=MQTTConnectProps)

    def __str__(self) -> str:
        attrs = [
            f"client_id={self.client_id}",
            f"keep_alive={self.keep_alive}",
            f"protocol_version={self.protocol_version}",
            f"clean_start={self.clean_start}",
            f"username={self.username}",
            f"password={str(len(self.password)) + 'B' if self.password else None}",
            f"will_topic={self.will_topic}",
            f"will_payload={str(len(self.will_payload)) + 'B' if self.will_payload else None}",
            f"will_qos={self.will_qos}",
            f"will_retain={self.will_retain}",
            f"will_props={self.will_props}",
            f"properties={self.properties}",
        ]
        return f"CONNECT[{', '.join(attrs)}]"

    def encode(self) -> bytes:
        connect_flags = (
            (self.clean_start << 1) |
            (self.will_qos << 3) |
            (self.will_retain << 5)
        )

        payload = bytearray()
        payload.extend(encode_string(self.client_id))
        if self.will_topic:
            payload.extend(self.will_props.encode() + encode_string(self.will_topic) + encode_binary(self.will_payload))
            connect_flags |= 0x04
        if self.username is not None:
            payload.extend(encode_string(self.username))
            connect_flags |= 0x80
        if self.password is not None:
            payload.extend(encode_binary(self.password))
            connect_flags |= 0x40

        data = b"".join((
            encode_binary(b"MQTT"),
            encode_uint8(self.protocol_version),
            encode_uint8(connect_flags),
            encode_uint16(self.keep_alive),
            self.properties.encode(),
            payload,
        ))
        head = HEAD_CONNECT
        length = encode_varint(len(data))
        encoded = bytearray()
        encoded.append(head)
        encoded.extend(length)
        encoded.extend(data)
        return bytes(encoded)

    @classmethod
    def decode(cls, flags: int, data: memoryview) -> MQTTConnectPacket:
        if flags != 0:
            raise MQTTError(f"Invalid flags, expected 0 but got {flags}", MQTTReasonCode.MalformedPacket)

        offset = 0
        protocol_name, sz = decode_binary(data)
        offset += sz

        if protocol_name != b"MQTT":
            raise MQTTError("Invalid protocol name", MQTTReasonCode.ProtocolError)

        protocol_version, sz = decode_uint8(data[offset:])
        offset += sz

        if protocol_version != 5:
            raise MQTTError(f"Invalid protocol version, expected 5 but got {protocol_version}", MQTTReasonCode.UnsupportedProtocolVersion)

        connect_flags, sz = decode_uint8(data[offset:])
        offset += sz

        clean_start = connect_flags & 0x02 == 2
        will_flag = connect_flags & 0x04 == 4
        will_qos = connect_flags >> 3 & 0x03
        will_retain = connect_flags & 0x20 == 32
        password_flag = connect_flags & 0x40 == 64
        username_flag = connect_flags & 0x80 == 128

        keep_alive, sz = decode_uint16(data[offset:])
        offset += sz

        props, sz = MQTTConnectProps.decode(data[offset:])
        offset += sz

        client_id, sz = decode_string(data[offset:])
        offset += sz

        if will_flag:
            will_props, sz = MQTTWillProps.decode(data[offset:])
            offset += sz
            will_topic, sz = decode_string(data[offset:])
            offset += sz
            will_payload, sz = decode_binary(data[offset:])
            offset += sz
        else:
            will_props = MQTTWillProps()
            will_topic = ""
            will_payload = b""

        if username_flag:
            username, sz = decode_string(data[offset:])
            offset += sz
        else:
            username = None

        if password_flag:
            password, sz = decode_binary(data[offset:])
            offset += sz
        else:
            password = None

        return MQTTConnectPacket(
            client_id,
            keep_alive,
            protocol_version,
            clean_start=clean_start,
            will_props=will_props,
            will_topic=will_topic,
            will_payload=will_payload,
            will_qos=will_qos,
            will_retain=will_retain,
            password=password,
            username=username,
            properties=props,
        )


@dataclass(match_args=True, slots=True)
class MQTTConnAckPacket(MQTTPacket):
    packet_type = MQTTPacketType.CONNACK
    props_type = MQTTConnAckProps
    reason_code: MQTTReasonCode = MQTTReasonCode.Success
    session_present: bool = False
    properties: MQTTConnAckProps = field(default_factory=MQTTConnAckProps)

    def __str__(self) -> str:
        attrs = [
            f"reason_code={hex(self.reason_code)}",
            f"session_present={self.session_present}",
            f"properties={self.properties}",
        ]
        return f"CONNACK[{', '.join(attrs)}]"

    def encode(self) -> bytes:
        head = HEAD_CONNACK
        data = encode_bool(self.session_present) + encode_uint8(self.reason_code.value) + self.properties.encode()
        length = encode_varint(len(data))
        return b"".join((bytes([head]), length, data))

    @classmethod
    def decode(cls, flags: int, data: memoryview) -> MQTTConnAckPacket:
        if flags != 0:
            raise MQTTError(f"Invalid flags, expected 0 but got {flags}", MQTTReasonCode.MalformedPacket)

        session_present, _ = decode_bool(data[0:])
        reason_code = MQTTReasonCode(decode_uint8(data[1:])[0])
        props, _ = MQTTConnAckProps.decode(data[2:])
        return MQTTConnAckPacket(reason_code, session_present, properties=props)


@dataclass(match_args=True, slots=True)
class MQTTDisconnectPacket(MQTTPacket):
    packet_type = MQTTPacketType.DISCONNECT
    props_type = MQTTDisconnectProps
    reason_code: MQTTReasonCode = MQTTReasonCode.Success
    properties: MQTTDisconnectProps = field(default_factory=MQTTDisconnectProps)

    def __str__(self) -> str:
        attrs = [
            f"reason_code={hex(self.reason_code)}",
            f"properties={self.properties}",
        ]
        return f"DISCONNECT[{', '.join(attrs)}]"

    def encode(self) -> bytes:
        # If the reason code is success and there are no properties, the packet can be empty.
        has_props = len(self.properties) > 0
        if self.reason_code == MQTTReasonCode.Success and not has_props:
            return HEAD_DISCONNECT.to_bytes(1, "big") + b"\x00"
        encoded = bytearray()
        encoded.append(HEAD_DISCONNECT)
        props = self.properties.encode() if has_props else b""
        length = 1 + len(props)
        encoded.extend(encode_varint(length))
        encoded.append(self.reason_code.value)
        if has_props:
            encoded.extend(props)
        return bytes(encoded)

    @classmethod
    def decode(cls, flags: int, data: memoryview) -> MQTTDisconnectPacket:
        if flags != 0:
            raise MQTTError(f"Invalid flags, expected 0 but got {flags}", MQTTReasonCode.MalformedPacket)
        if len(data) == 0:
            # An empty packet means success with no properties.
            return MQTTDisconnectPacket()
        reason_code = MQTTReasonCode(decode_uint8(data)[0])
        if len(data) == 1:
            # If the packet only contains the reason code, it has no properties.
            return MQTTDisconnectPacket(reason_code)
        props, _ = MQTTDisconnectProps.decode(data[1:])
        return MQTTDisconnectPacket(reason_code, properties=props)



from __future__ import annotations

from collections import deque
from dataclasses import dataclass, field
import socket
import ssl
from typing import Callable, TypeAlias

from .address import Address
from .decoder import IncrementalDecoder
from .keepalive import KeepAlive
from .timeout import Timeout
from .ws_decoder import WebsocketDecoder
from ..mqtt_spec import MQTTReasonCode
from ..packet import (
    MQTTPacket,
    MQTTConnAckPacket,
    MQTTDisconnectPacket,
    MQTTAuthPacket,
    MQTTPublishPacket,
    MQTTPubAckPacket,
    MQTTPubRecPacket,
    MQTTPubRelPacket,
    MQTTPubCompPacket,
    MQTTSubscribePacket,
    MQTTUnsubscribePacket,
    MQTTSubAckPacket,
    MQTTUnsubAckPacket,
)
from ..property import MQTTConnectProps, MQTTWillProps


ReceivablePacketT: TypeAlias = (
    MQTTConnAckPacket |
    MQTTPublishPacket |
    MQTTPubAckPacket |
    MQTTPubRecPacket |
    MQTTPubRelPacket |
    MQTTPubCompPacket |
    MQTTSubAckPacket |
    MQTTUnsubAckPacket |
    MQTTAuthPacket |
    MQTTDisconnectPacket
)
SendablePacketT: TypeAlias = (
    MQTTPublishPacket |
    MQTTPubAckPacket |
    MQTTPubRecPacket |
    MQTTPubRelPacket |
    MQTTPubCompPacket |
    MQTTSubscribePacket |
    MQTTUnsubscribePacket |
    MQTTAuthPacket
)
ConnectionReadCallback: TypeAlias = Callable[[ReceivablePacketT], None]


@dataclass(slots=True, match_args=True, frozen=True)
class ConnectParams:
    """Parameters for the MQTT connection."""
    address: Address = field(default_factory=Address)
    client_id: str = ""
    connect_timeout: float | None = None
    reconnect_delay: int = 0
    keepalive_interval: int = 0
    tcp_nodelay: bool = True
    tls_context: ssl.SSLContext | None = None
    tls_hostname: str = ""
    protocol_version: int = 5
    clean_start: bool = False
    username: str | None = None
    password: bytes | None = None
    will_topic: str = ""
    will_payload: bytes = b""
    will_qos: int = 0
    will_retain: bool = False
    will_properties: MQTTWillProps = field(default_factory=MQTTWillProps)
    connect_properties: MQTTConnectProps = field(default_factory=MQTTConnectProps)


@dataclass(kw_only=True, slots=True)
class StateData:
    """State data for the connection.

    This should contain any attributes needed by multiple states.

    The data in this class should never be accessed from outside the state methods."""
    sock: socket.socket | ssl.SSLSocket = field(init=False, default_factory=socket.socket)
    disconnect_rc: MQTTReasonCode | None = field(init=False, default=None)
    keepalive: KeepAlive = field(init=False, default_factory=KeepAlive)
    timeout: Timeout = field(init=False, default_factory=Timeout)
    decoder: IncrementalDecoder = field(init=False, default_factory=IncrementalDecoder)
    connack: MQTTConnAckPacket | None = field(init=False, default=None)
    open_called: bool = field(init=False, default=False)
    ws_decoder: WebsocketDecoder = field(init=False, default_factory=WebsocketDecoder)
    ws_nonce: str = field(init=False, default="")
    ws_handshake_buffer: bytearray = field(init=False, default_factory=bytearray)
    write_buffer: bytearray = field(init=False, default_factory=bytearray)


@dataclass(slots=True, kw_only=True)
class StateEnvironment:
    """State environment for the connection.

    Data in this class is shared with the outside world."""
    packet_callback: ConnectionReadCallback
    packet_buffer: deque[MQTTPacket] = field(init=False, default_factory=deque)



from typing import Any

import pytest

from ohmqtt.error import MQTTError
from ohmqtt.mqtt_spec import MQTTReasonCode
from ohmqtt.serialization import (
    encode_bool,
    decode_bool,
    encode_uint8,
    decode_uint8,
    encode_uint16,
    decode_uint16,
    encode_uint32,
    decode_uint32,
    encode_string,
    decode_string,
    encode_string_pair,
    decode_string_pair,
    encode_binary,
    decode_binary,
    encode_varint,
    decode_varint,
)


def _pull_hex_view(hex_str: str) -> memoryview:
    """Helper to pull input data from test data."""
    return memoryview(bytes.fromhex(hex_str))


def test_encode_bool(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        assert encode_bool(case["input"]) == bytes.fromhex(case["output"])
        decoded, sz = decode_bool(_pull_hex_view(case["output"]))
        assert sz == 1

        assert decoded == case["input"]


def test_decode_bool_errors(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        with pytest.raises(MQTTError):
            decode_bool(_pull_hex_view(case["input"]))


def test_encode_uint8(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        assert encode_uint8(case["input"]) == bytes.fromhex(case["output"])
        decoded, sz = decode_uint8(_pull_hex_view(case["output"]))
        assert sz == 1

        assert decoded == case["input"]


def test_decode_uint8_errors(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        with pytest.raises(MQTTError):
            decode_uint8(_pull_hex_view(case["input"]))


def test_encode_uint16(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        assert encode_uint16(case["input"]) == bytes.fromhex(case["output"])
        decoded, sz = decode_uint16(_pull_hex_view(case["output"]))
        assert sz == 2

        assert decoded == case["input"]


def test_decode_uint16_errors(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        with pytest.raises(MQTTError):
            decode_uint16(_pull_hex_view(case["input"]))


def test_encode_uint32(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        assert encode_uint32(case["input"]) == bytes.fromhex(case["output"])
        decoded, sz = decode_uint32(_pull_hex_view(case["output"]))
        assert sz == 4

        assert decoded == case["input"]


def test_decode_uint32_errors(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        with pytest.raises(MQTTError):
            decode_uint32(_pull_hex_view(case["input"]))


def test_encode_string(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        encoded = encode_string(case["input"])
        assert encoded == bytes.fromhex(case["output"]), encoded.hex()
        decoded, sz = decode_string(memoryview(encoded))
        assert sz == len(encoded)
        assert decoded == case["input"]


def test_decode_string(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        decoded, sz = decode_string(_pull_hex_view(case["input"]))
        assert sz <= len(_pull_hex_view(case["input"]))
        assert decoded == case["output"], case["input"]


def test_decode_string_errors(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        with pytest.raises(MQTTError):
            decode_string(_pull_hex_view(case["input"]))


def test_decode_string_surrogates() -> None:
    # A separate test to ensure invalid surrogate characters are handled correctly by Python.
    for n in range(0xD800, 0xDFFF+1):
        s = chr(n)
        x = s.encode("utf-8", errors="surrogatepass")
        y = memoryview(len(x).to_bytes(2, byteorder="big") + x)
        with pytest.raises(MQTTError) as exc:
            decode_string(y)
        assert exc.value.reason_code == MQTTReasonCode.MalformedPacket


def test_encode_string_pair(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        pair = tuple(case["input"])
        encoded = encode_string_pair(pair)
        assert encoded == bytes.fromhex(case["output"]), encoded.hex()
        decoded, sz = decode_string_pair(memoryview(encoded))
        assert sz == len(encoded)
        assert decoded == pair


def test_decode_string_pair_errors(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        with pytest.raises(MQTTError):
            decode_string_pair(_pull_hex_view(case["input"]))


def test_encode_binary(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        input_data = bytes.fromhex(case["input"])
        encoded = encode_binary(input_data)
        assert encoded == bytes.fromhex(case["output"]), encoded.hex()
        decoded, sz = decode_binary(memoryview(encoded))
        assert sz == len(encoded)
        assert decoded == input_data


def test_decode_binary_errors(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        with pytest.raises(MQTTError):
            decode_binary(_pull_hex_view(case["input"]))


def test_encode_varint(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        assert encode_varint(case["input"]) == bytes.fromhex(case["output"])


def test_encode_varint_limits(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        with pytest.raises(ValueError):
            encode_varint(case["input"])


def test_decode_varint(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        decoded, sz = decode_varint(_pull_hex_view(case["input"]))
        assert sz <= len(bytes.fromhex(case["input"]))
        assert decoded == case["output"]


def test_decode_varint_limits(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        with pytest.raises(MQTTError):
            decode_varint(_pull_hex_view(case["input"]))



from contextlib import suppress
import sqlite3
from threading import Condition
from typing import Final
import weakref

from .base import LostMessageError, Persistence, RenderedPacket
from ..error import MQTTError
from ..handles import PublishHandle
from ..logger import get_logger
from ..mqtt_spec import MAX_PACKET_ID, MQTTQoS
from ..packet import (
    MQTTPublishPacket,
    MQTTPubAckPacket,
    MQTTPubRecPacket,
    MQTTPubRelPacket,
    MQTTPubCompPacket,
)
from ..property import MQTTPublishProps
from ..topic_alias import AliasPolicy

logger: Final = get_logger("persistence.sqlite")

SCHEMA_VERSION: Final[int] = 0


class SchemaError(Exception):
    """Database schema error."""


class SQLitePersistence(Persistence):
    """SQLite persistence for MQTT messages.

    This class provides a SQLite-based persistence layer for MQTT messages.
    """
    __slots__ = ("_cond", "_conn", "_cursor", "_db_path", "_handles")
    def __init__(self, db_path: str, *, db_fast: bool = False) -> None:
        self._cond = Condition()
        self._handles: weakref.WeakValueDictionary[int, PublishHandle] = weakref.WeakValueDictionary({})
        self._db_path = db_path
        self._conn = sqlite3.connect(self._db_path, check_same_thread=False)
        self._cursor = self._conn.cursor()
        if db_fast:
            self._cursor.executescript(
                """
                PRAGMA journal_mode = WAL;
                PRAGMA synchronous = NORMAL;
                """
            )
        self._cursor.execute("PRAGMA user_version")
        current_version = self._cursor.fetchone()[0]
        if current_version != SCHEMA_VERSION:
            raise SchemaError(f"Database version {current_version} does not match library version {SCHEMA_VERSION}")
        self._create_tables()

    def __del__(self) -> None:
        self.close()

    def __len__(self) -> int:
        with self._cond:
            self._cursor.execute(
                """
                SELECT COUNT(*) FROM messages
                """
            )
            row = self._cursor.fetchone()
            return int(row[0])

    def _create_tables(self) -> None:
        """Create the necessary tables in the SQLite database."""
        with self._cond:
            self._cursor.execute(f"PRAGMA user_version = {SCHEMA_VERSION}")
            self._cursor.executescript(
                """
                BEGIN;
                CREATE TABLE IF NOT EXISTS messages (
                    id INTEGER PRIMARY KEY,
                    topic TEXT NOT NULL,
                    payload BLOB NOT NULL,
                    qos INTEGER NOT NULL,
                    retain INTEGER NOT NULL,
                    properties BLOB,
                    alias_policy INTEGER NOT NULL,
                    dup INTEGER DEFAULT 0,
                    received INTEGER DEFAULT 0,
                    packet_id INTEGER UNIQUE DEFAULT NULL,
                    inflight INTEGER DEFAULT 0
                ) STRICT;
                CREATE TABLE IF NOT EXISTS received (
                    packet_id INTEGER PRIMARY KEY
                ) STRICT;
                CREATE TABLE IF NOT EXISTS client_id (
                    id INTEGER PRIMARY KEY CHECK (id = 0),
                    client_id TEXT DEFAULT NULL
                ) STRICT;
                INSERT OR IGNORE INTO client_id (id) VALUES (0);
                COMMIT;
                """
            )

    def _get_client_id(self) -> str:
        """Get the client ID from the database."""
        with self._cond:
            self._cursor.execute(
                """
                SELECT client_id FROM client_id WHERE id = 0
                """
            )
            row = self._cursor.fetchone()
            return str(row[0])

    def _set_client_id(self, client_id: str) -> None:
        """Set the client ID in the database."""
        with self._cond:
            self._cursor.execute(
                """
                UPDATE client_id SET client_id = ? WHERE id = 0
                """,
                (client_id,),
            )
            self._conn.commit()

    def add(
        self,
        topic: str,
        payload: bytes,
        qos: MQTTQoS,
        retain: bool,
        properties: MQTTPublishProps,
        alias_policy: AliasPolicy,
    ) -> PublishHandle:
        if alias_policy == AliasPolicy.ALWAYS:
            raise ValueError("AliasPolicy must not be ALWAYS for retained messages.")
        with self._cond:
            properties_blob = properties.encode() if len(properties) > 0 else None
            self._cursor.execute(
                """
                INSERT INTO messages (topic, payload, qos, retain, properties, alias_policy)
                VALUES (?, ?, ?, ?, ?, ?)
                """,
                (topic, payload, qos.value, int(retain), properties_blob, alias_policy.value),
            )
            self._conn.commit()
            message_id = self._cursor.lastrowid
            assert message_id is not None
            handle = PublishHandle(self._cond)
            self._handles[message_id] = handle
            return handle

    def get(self, count: int) -> list[int]:
        with self._cond:
            self._cursor.execute(
                """
                SELECT id FROM messages
                WHERE inflight = 0
                ORDER BY id ASC
                LIMIT ?
                """,
                (count,),
            )
            rows = self._cursor.fetchall()
            return [row[0] for row in rows]

    def ack(self, packet: MQTTPubAckPacket | MQTTPubRecPacket | MQTTPubCompPacket) -> None:
        packet_id = packet.packet_id
        rc = packet.reason_code
        with self._cond:
            self._cursor.execute(
                """
                SELECT id, qos, received FROM messages WHERE packet_id = ?
                """,
                (packet_id,),
            )
            row = self._cursor.fetchone()
            if row is None:
                raise ValueError(f"Unknown packet_id: {packet_id}")
            message_id, qos, received = row
            if received or qos == 1 or rc.is_error():
                # If the message is QoS 1, we need to delete it from the store.
                self._cursor.execute(
                    """
                    DELETE FROM messages WHERE id = ?
                    """,
                    (message_id,),
                )
                handle = self._handles.pop(message_id, None)
                if handle is not None:
                    handle.ack = packet
                    if rc.is_error():
                        handle.exc = MQTTError("Error code in acknowledgement packet", rc)
                    self._cond.notify_all()
            else:
                # If the message is QoS 2, we need to mark it as received.
                self._cursor.execute(
                    """
                    UPDATE messages SET inflight = 0, received = 1 WHERE id = ?
                    """,
                    (message_id,),
                )
            self._conn.commit()

    def check_rec(self, packet: MQTTPublishPacket) -> bool:
        if packet.qos != MQTTQoS.Q2:
            raise ValueError("Not a QoS 2 PUBLISH packet")
        with self._cond:
            self._cursor.execute(
                """
                SELECT packet_id FROM received WHERE packet_id = ?
                """,
                (packet.packet_id,),
            )
            row = self._cursor.fetchone()
            if row is not None:
                logger.debug("Received duplicate QoS 2 packet with ID %d", packet.packet_id)
                return False
            return True

    def set_rec(self, packet: MQTTPublishPacket) -> None:
        if packet.qos != MQTTQoS.Q2:
            raise ValueError("Not a QoS 2 PUBLISH packet")
        with self._cond:
            self._cursor.execute(
                """
                INSERT INTO received (packet_id) VALUES (?)
                """,
                (packet.packet_id,),
            )
            self._conn.commit()

    def rel(self, packet: MQTTPubRelPacket) -> None:
        with self._cond:
            self._cursor.execute(
                """
                DELETE FROM received WHERE packet_id = ?
                """,
                (packet.packet_id,),
            )
            self._conn.commit()

    def _generate_packet_id(self, message_id: int) -> int:
        """Generate a unique packet ID for the message."""
        packet_id = message_id
        while packet_id > MAX_PACKET_ID:
            packet_id -= MAX_PACKET_ID
        return packet_id

    def render(self, message_id: int) -> RenderedPacket:
        with self._cond:
            self._cursor.execute(
                """
                SELECT topic, payload, qos, retain, properties, dup, received, packet_id, alias_policy, (SELECT MIN(id) FROM messages WHERE inflight = 0)
                FROM messages
                WHERE id = ?
                """,
                (message_id,),
            )
            row = self._cursor.fetchone()
            if row is None:
                raise KeyError(f"Message ID {message_id} not found in persistence store.")
            topic, payload, qos, retain, properties_blob, dup, received, packet_id, alias_policy, min_id = row
            if message_id != min_id:
                raise ValueError(f"Message {message_id} is not next in queue.")
            if properties_blob is not None:
                properties_view = memoryview(properties_blob)
                properties, _ = MQTTPublishProps.decode(properties_view)
            else:
                properties = MQTTPublishProps()
            if packet_id is None:
                packet_id = self._generate_packet_id(message_id)
            qos = MQTTQoS(qos)
            packet: MQTTPublishPacket | MQTTPubRelPacket
            if received:
                alias_policy = AliasPolicy.NEVER
                packet = MQTTPubRelPacket(packet_id=packet_id)
            else:
                alias_policy = AliasPolicy(alias_policy)
                packet = MQTTPublishPacket(
                    topic=topic,
                    payload=payload,
                    packet_id=packet_id,
                    qos=qos,
                    retain=bool(retain),
                    properties=properties,
                    dup=dup,
                )
            self._cursor.execute(
                """
                UPDATE messages SET inflight = 1, packet_id = ? WHERE id = ?
                """,
                (packet_id, message_id),
            )
            self._conn.commit()
            return RenderedPacket(packet, alias_policy)

    def _reset_inflight(self) -> None:
        """Clear inflight status of all messages."""
        with self._cond:
            self._cursor.executescript(
                """
                BEGIN;
                UPDATE messages SET dup = 1 WHERE inflight = 1;
                UPDATE messages SET inflight = 0;
                COMMIT;
                """
            )

    def clear(self) -> None:
        with self._cond:
            self._cursor.executescript(
                """
                BEGIN;
                DELETE FROM messages;
                DELETE FROM received;
                COMMIT;
                """
            )
            if self._handles:
                for handle in self._handles.values():
                    handle.exc = LostMessageError("Message lost from persistence store")
                self._cond.notify_all()
            self._handles.clear()

    def open(self, client_id: str, clear: bool = False) -> None:
        logger.debug("Opening SQLite persistence with client ID: %s clear=%s", client_id, clear)
        with self._cond:
            if clear or client_id != self._get_client_id():
                logger.debug("Clearing SQLite persistence for client ID: %s", client_id)
                self._set_client_id(client_id)
                self.clear()
            else:
                self._reset_inflight()

    def close(self) -> None:
        # Ignore "already closed" errors while closing resources.
        with suppress(sqlite3.ProgrammingError):
            self._cursor.close()
        with suppress(sqlite3.ProgrammingError):
            self._conn.close()



from __future__ import annotations

from dataclasses import dataclass
import threading
from typing import Final, Sequence

from .selector import InterruptibleSelector
from .states.base import FSMState
from .timeout import Timeout
from .types import ConnectParams, StateData, StateEnvironment
from ..logger import get_logger

logger: Final = get_logger("connection.fsm")


class InvalidStateError(Exception):
    """Exception raised when an operation is performed in an invalid state."""


@dataclass(slots=True, init=False)
class FSM:
    """Threadsafe Finite State Machine."""
    env: StateEnvironment
    previous_state: type[FSMState]
    requested_state: type[FSMState]
    state: type[FSMState]
    error_state: type[FSMState]
    lock: threading.RLock
    cond: threading.Condition
    selector: InterruptibleSelector
    params: ConnectParams
    _state_changed: bool
    _state_requested: bool
    _state_data: StateData

    def __init__(self, env: StateEnvironment, init_state: type[FSMState], error_state: type[FSMState]) -> None:
        self.env = env
        self.previous_state = init_state
        self.requested_state = init_state
        self.state = init_state
        self.error_state = error_state
        self.lock = threading.RLock()
        self.cond = threading.Condition(self.lock)
        self.selector = InterruptibleSelector(self.lock)
        self.params = ConnectParams()
        self._state_changed = True
        self._state_requested = False
        self._state_data = StateData()

    def set_params(self, params: ConnectParams) -> None:
        """Set the connection parameters."""
        with self.lock:
            self.params = params

    def get_state(self) -> type[FSMState]:
        """Get the current state."""
        with self.lock:
            return self.state

    def wait_for_state(self, states: Sequence[type[FSMState]], timeout: float | None = None) -> bool:
        """Wait for a specific state to be reached.

        :return: True if the state is reached, False if the timeout is reached."""
        with self.cond:
            if self.state in states:
                return True
            return self.cond.wait_for(lambda: self.state in states, timeout)

    def change_state(self, state: type[FSMState]) -> None:
        """Change to a new state.

        This method must only be called from within a state."""
        with self.cond:
            if state is self.state:
                return
            if not self.state.can_transition_to(state):
                raise InvalidStateError(f"Cannot transition from {self.state.__name__} to {state.__name__}")
            self.previous_state = self.state
            self.state = state
            self._state_changed = True
            self.cond.notify_all()

    def request_state(self, state: type[FSMState]) -> None:
        """Request a state change from outside the FSM."""
        with self.cond:
            if not self.state.can_transition_to(state) or not state.can_request_from(self.state):
                logger.debug("Ignoring invalid request to change from %s to %s", self.state.__name__, state.__name__)
                return
            logger.debug("Requesting state change from %s to %s", self.state.__name__, state.__name__)
            self.requested_state = state
            self._state_requested = True
            self.cond.notify_all()

    def _handle_exception(self, exc: Exception) -> None:
        """Handle an exception by transitioning to the error state.

        :raises InvalidStateError: The error state cannot be transitioned to from the current state."""
        with self.cond:
            if self.state != self.error_state:
                if not self.state.can_transition_to(self.error_state):
                    logger.exception("Unhandled exception in FSM loop, cannot transition to %s, exploding", self.error_state.__name__)
                    raise InvalidStateError(f"Cannot transition to error state {self.error_state.__name__} from {self.state.__name__}") from exc
                logger.exception("Unhandled exception in FSM loop, going to %s", self.error_state.__name__)
                self.previous_state = self.state
                self.state = self.error_state
                self._state_changed = True
                self._state_requested = False
                self.cond.notify_all()

    def loop_once(self, max_wait: float | None = 0.0) -> bool:
        """Do the current state.

        State transition will be run if needed.

        :return: True if the state is finished and the calling thread should wait for a change."""
        try:
            with self.cond:
                # Consume state change requests.
                if self._state_requested:
                    if not self.state.can_transition_to(self.requested_state) or not self.requested_state.can_request_from(self.state):
                        logger.debug("Ignoring invalid request to change from %s to %s", self.state.__name__, self.requested_state.__name__)
                        self._state_requested = False
                        self.requested_state = self.state
                    else:
                        logger.debug("Handling request to change from %s to %s", self.state.__name__, self.requested_state.__name__)
                        self.previous_state = self.state
                        self.state = self.requested_state
                        self._state_requested = False
                        self._state_changed = True
                # Consume state changes, either from a state or from a request.
                if self._state_changed:
                    self._state_changed = False
                    logger.debug("Entering state %s", self.state.__name__)
                    self.state.enter(self, self._state_data, self.env, self.params)
                    return False  # Run the state on the next loop, unless it has changed.

            # Run the state.  Do not run this while holding the lock, as it may block.
            return self.state.handle(self, self._state_data, self.env, self.params, max_wait)
        except Exception as exc:
            # If there is an unhandled exception anywhere in the loop, try to go to the error state.
            self._handle_exception(exc)
            raise

    def loop_until_state(self, targets: Sequence[type[FSMState]], timeout: float | None = None) -> bool:
        """Run the state machine until a specific state(s) has been entered.

        :return: True if a target state is reached, False if another final state was finished or timeout reached."""
        to = Timeout(timeout)

        def _state_unreachable() -> bool:
            """Check that we can not reach the target state from where we are.

            Must be called with the self.cond lock held.

            This method may block while waiting for a state change, state request, or timeout."""
            return (
                state_done and (
                    self.state.is_final_state() or
                    (not (self._state_changed or self._state_requested) and not self.cond.wait(timeout=to.get_timeout()))
                )
            )

        while True:
            state_done = self.loop_once(max_wait=to.get_timeout())
            with self.cond:
                if self.state in targets and not self._state_changed:
                    # We are in a target state and the state has been entered.
                    return True
                if to.exceeded() or _state_unreachable():
                    return False



#!/usr/bin/env python3
"""This example demonstrates a simple RPC server.

It listens for incoming RPC requests on a specific topic and responds with the result of the RPC call.

The ResponseTopic property is used by the requestor to specify the topic to which the response should be sent.

See the "rpc_caller" example for the request side of this RPC implementation.

This also demonstrates running the client loop in the main thread."""

import argparse

from ohmqtt import Client, MQTTPublishPacket, MQTTPublishProps, MQTTQoS


class RPCServer:
    """A simple stateless RPC server."""
    def handle_request(self, client: Client, msg: MQTTPublishPacket) -> None:
        """Handle incoming RPC requests."""
        print(f"*** Received RPC request: {msg!s}")

        # Find the response topic in the message properties.
        if not msg.properties.ResponseTopic:
            print("Request was missing required response topic property")
            return
        response_topic = msg.properties.ResponseTopic

        response_props = MQTTPublishProps()
        # If the request includes correlation data, send it back in the response.
        if msg.properties.CorrelationData is not None:
            response_props.CorrelationData = msg.properties.CorrelationData

        # Simulate some processing.
        response = f"This is a good day for {msg.payload.decode()}"

        # Send the response back to the specified topic.
        client.publish(response_topic, response.encode(), qos=MQTTQoS.Q2, properties=response_props)


def main(args: argparse.Namespace) -> None:
    rpc_server = RPCServer()
    client = Client()
    client.connect(args.address)
    print("*** Waiting for connection...")
    assert client.loop_until_connected(timeout=5.0), "Timeout waiting for connection"

    client.subscribe("ohmqtt/examples/rpc/request", rpc_server.handle_request)
    print("*** Waiting for RPC requests...")

    try:
        client.loop_forever()  # Wait indefinitely for incoming messages.
    except KeyboardInterrupt:
        print("\n*** Shutting down RPC server...")


if __name__ == "__main__":
    from .args import parser
    args = parser.parse_args()
    main(args)



import socket
import ssl
from unittest.mock import Mock

import pytest
from pytest_mock import MockerFixture

from ohmqtt.connection.decoder import IncrementalDecoder, ClosedSocketError
from ohmqtt.error import MQTTError
from ohmqtt.mqtt_spec import MQTTReasonCode, MQTTQoS
from ohmqtt.packet import MQTTPublishPacket


@pytest.fixture
def mock_socket(mocker: MockerFixture) -> Mock:
    """Return a mock socket for testing."""
    return mocker.Mock(spec=socket.socket)  # type: ignore[no-any-return]


def test_decoder_defaults() -> None:
    """reset() should set the decoder to its default state."""
    decoder = IncrementalDecoder()
    decoder2 = IncrementalDecoder()
    decoder2.reset()
    assert decoder.head == decoder2.head
    assert decoder.length == decoder2.length
    assert decoder.data == decoder2.data


@pytest.mark.parametrize("exc", [BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError])
def test_decoder_drip_feed(exc: Exception, mock_socket: Mock) -> None:
    """Feed the decoder one byte at a time."""
    decoder = IncrementalDecoder()
    # Use a packet with a payload of 255 bytes to ensure we have to read
    # multiple bytes for the length varint.
    in_packet = MQTTPublishPacket(topic="topic", payload=b"x" * 255, qos=MQTTQoS.Q1, packet_id=66)
    data = in_packet.encode()

    # Setup mock to return one byte at a time
    for i in range(len(data) - 1):
        mock_socket.recv.side_effect = [data[i:i+1], exc]
        assert decoder.decode(mock_socket) is None

    # Final byte completes the packet
    mock_socket.recv.side_effect = [data[-1:]]
    out_packet = decoder.decode(mock_socket)
    assert out_packet == in_packet


@pytest.mark.parametrize("exc", [BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError])
@pytest.mark.parametrize("closer", [b"", OSError])
@pytest.mark.parametrize("available_bytes", [n for n in range(6)])
def test_decoder_drip_partial_closures(exc: Exception, closer: bytes | Exception,
                                       available_bytes: int, mock_socket: Mock) -> None:
    """Feed the decoder partial data, then close the socket."""
    decoder = IncrementalDecoder()
    # Use a packet with a payload of 255 bytes to ensure we have to read
    # multiple bytes for the length varint.
    in_packet = MQTTPublishPacket(topic="topic", payload=b"x" * 255, qos=MQTTQoS.Q1, packet_id=66)
    data = in_packet.encode()

    # Send a few bytes then simulate closed socket
    if available_bytes > 0:
        mock_socket.recv.side_effect = [data[:available_bytes], exc]
    else:
        mock_socket.recv.side_effect = [exc]

    assert decoder.decode(mock_socket) is None

    mock_socket.recv.side_effect = [closer]
    with pytest.raises(ClosedSocketError):
        decoder.decode(mock_socket)


@pytest.mark.parametrize("bad_data", [
    b"\xf0\x80\x80\x80\x80",
    b"\xf0\xff\xff\xff\xff",
])
def test_decoder_bad_length(bad_data: bytes, mock_socket: Mock) -> None:
    """Feed the decoder a bad varint length."""
    decoder = IncrementalDecoder()
    mock_socket.recv.return_value = bad_data
    with pytest.raises(MQTTError, match=r"Varint overflow") as excinfo:
        decoder.decode(mock_socket)
    assert excinfo.value.reason_code == MQTTReasonCode.MalformedPacket


def test_decoder_empty_reads(mock_socket: Mock) -> None:
    """Feed the decoder empty reads."""
    decoder = IncrementalDecoder()
    # Use a packet with a payload of 255 bytes to ensure we have to read
    # multiple bytes for the length varint.
    in_packet = MQTTPublishPacket(topic="topic", payload=b"x" * 255, qos=MQTTQoS.Q1, packet_id=66)
    data = in_packet.encode()

    # Empty read on first byte.
    mock_socket.recv.side_effect = [b""]
    with pytest.raises(ClosedSocketError):
        decoder.decode(mock_socket)
    decoder.reset()

    # Empty read on first varint byte.
    mock_socket.recv.side_effect = [data[:1], b""]
    with pytest.raises(ClosedSocketError):
        decoder.decode(mock_socket)
    decoder.reset()

    # Empty read on second varint byte.
    mock_socket.recv.side_effect = [data[:1], data[1:2], b""]
    with pytest.raises(ClosedSocketError):
        decoder.decode(mock_socket)
    decoder.reset()

    # Empty read on first byte of content.
    mock_socket.recv.side_effect = [data[:1], data[1:2], data[2:3], b""]
    with pytest.raises(ClosedSocketError):
        decoder.decode(mock_socket)
    decoder.reset()

    # Empty reads on each content byte.
    for n in range(3, len(data)):
        mock_socket.recv.side_effect = [data[:1], data[1:2], data[2:3], data[3:n], b""]
        with pytest.raises(ClosedSocketError):
            decoder.decode(mock_socket)
        decoder.reset()

    mock_socket.recv.side_effect = [data[:1], data[1:2], data[2:3], data[3:]]
    assert decoder.decode(mock_socket) == in_packet
    decoder.reset()



import logging

from ohmqtt.logger import get_logger, set_log_level


def test_logger_set_level() -> None:
    logger = get_logger("test")
    for level in (
        logging.DEBUG,
        logging.INFO,
        logging.WARNING,
        logging.ERROR,
        logging.CRITICAL,
    ):
        set_log_level(level)
        assert logger.getEffectiveLevel() == level



from __future__ import annotations

import threading
import time
from typing import Any

import pytest
from pytest_mock import MockerFixture

from ohmqtt.connection import Address
from ohmqtt.connection.fsm import FSM, InvalidStateError
from ohmqtt.connection.states.base import FSMState
from ohmqtt.connection.types import ConnectParams, StateData, StateEnvironment


class _TestError(Exception):
    """Custom exception for testing purposes."""


class MockState(FSMState):
    """Simulate a blocking state."""
    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment,
               params: ConnectParams, max_wait: float | None) -> bool:
        with fsm.cond:
            fsm.cond.wait(max_wait)
        return True

class MockStateA(MockState):
    """This is set up like initial state."""

class MockStateB(MockState):
    """This is an intermediate state."""

class MockStateC(MockState):
    """This state is final."""
    @classmethod
    def handle(cls, *args: Any, **kwargs: Any) -> bool:
        return True

class MockStateErr(MockState):
    """This state is just broken."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        raise _TestError("TEST Error in state enter")

    @classmethod
    def handle(cls, *args: Any, **kwargs: Any) -> bool:
        raise _TestError("TEST Error in state handler")


MockStateA.transitions_to = (MockStateB, MockStateC, MockStateErr)

MockStateB.transitions_to = (MockStateC, MockStateErr)
MockStateB.request_from = (MockStateA,)

MockStateC.request_from = (MockStateA, MockStateB)

MockStateErr.transitions_to = (MockStateC,)
MockStateErr.request_from = (MockStateA, MockStateB)


class EnvironmentCallbacks:
    """Container for StateEnvironment callbacks."""
    def __init__(self, mocker: MockerFixture) -> None:
        self.packet = mocker.Mock()


@pytest.fixture
def callbacks(mocker: MockerFixture) -> EnvironmentCallbacks:
    return EnvironmentCallbacks(mocker)


@pytest.fixture
def env(callbacks: EnvironmentCallbacks) -> StateEnvironment:
    """Fixture to create a StateEnvironment."""
    return StateEnvironment(
        packet_callback=callbacks.packet,
    )


def test_fsm_init(env: StateEnvironment) -> None:
    fsm = FSM(env=env, init_state=MockStateA, error_state=MockStateC)
    assert fsm.env == env
    assert fsm.previous_state == MockStateA
    assert fsm.requested_state == MockStateA
    assert fsm.state == MockStateA
    assert fsm.error_state == MockStateC


def test_fsm_props(env: StateEnvironment) -> None:
    fsm = FSM(env=env, init_state=MockStateA, error_state=MockStateC)

    params = ConnectParams(address=Address("test_address"))
    fsm.set_params(params)
    assert fsm.params == params

    assert fsm.get_state() == MockStateA

    with pytest.raises(TypeError):
        MockStateA()


@pytest.mark.parametrize("do_request", [True, False])
def test_fsm_wait_for_state(do_request: bool, env: StateEnvironment) -> None:
    fsm = FSM(env=env, init_state=MockStateA, error_state=MockStateC)
    assert fsm.wait_for_state([MockStateA], 0.001) is True
    assert fsm.wait_for_state([MockStateB, MockStateC], 0.001) is False

    start = threading.Event()
    def notifier() -> None:
        start.wait()
        if do_request:
            fsm.request_state(MockStateB)
        else:
            fsm.change_state(MockStateB)
        fsm.loop_once()

    thread = threading.Thread(target=notifier, daemon=True)
    thread.start()

    try:
        start.set()
        assert fsm.wait_for_state([MockStateB, MockStateC], 0.5) is True
    finally:
        thread.join(0.1)
        assert not thread.is_alive()


@pytest.mark.parametrize("do_request", [True, False])
def test_fsm_loop_until_state(do_request: bool, env: StateEnvironment) -> None:
    fsm = FSM(env=env, init_state=MockStateA, error_state=MockStateC)

    start = threading.Event()
    def notifier() -> None:
        start.wait()
        time.sleep(0.1)
        if do_request:
            fsm.request_state(MockStateB)
        else:
            fsm.change_state(MockStateB)

    thread = threading.Thread(target=notifier, daemon=True)
    thread.start()

    try:
        start.set()
        assert fsm.loop_until_state([MockStateB]) is True
    finally:
        thread.join(0.1)
        assert not thread.is_alive()


def test_fsm_loop_until_state_timeout(env: StateEnvironment) -> None:
    fsm = FSM(env=env, init_state=MockStateA, error_state=MockStateC)
    t0 = time.monotonic()
    assert fsm.loop_until_state([MockStateC], timeout=0.001) is False
    assert time.monotonic() - t0 < 0.1


@pytest.mark.parametrize("do_request", [True, False])
def test_fsm_loop_until_state_error(do_request: bool, env: StateEnvironment) -> None:
    fsm = FSM(env=env, init_state=MockStateA, error_state=MockStateC)

    start = threading.Event()
    def notifier() -> None:
        start.wait()
        time.sleep(0.1)
        if do_request:
            fsm.request_state(MockStateErr)
        else:
            fsm.change_state(MockStateErr)

    thread = threading.Thread(target=notifier, daemon=True)
    thread.start()

    try:
        start.set()
        with pytest.raises(_TestError):
            fsm.loop_until_state([MockStateB])
    finally:
        thread.join(0.1)
        assert not thread.is_alive()


@pytest.mark.parametrize("do_request", [True, False])
def test_fsm_loop_until_state_final(do_request: bool, env: StateEnvironment) -> None:
    fsm = FSM(env=env, init_state=MockStateA, error_state=MockStateC)

    start = threading.Event()
    def notifier() -> None:
        start.wait()
        time.sleep(0.1)
        if do_request:
            fsm.request_state(MockStateC)
        else:
            fsm.change_state(MockStateC)

    thread = threading.Thread(target=notifier, daemon=True)
    thread.start()

    try:
        start.set()
        assert fsm.loop_until_state([MockStateB]) is False
    finally:
        thread.join(0.1)
        assert not thread.is_alive()


def test_fsm_loop_once_error(env: StateEnvironment) -> None:
    fsm = FSM(env=env, init_state=MockStateErr, error_state=MockStateC)

    with pytest.raises(_TestError):
        fsm.loop_once()
    assert fsm.state == MockStateC
    assert fsm.previous_state == MockStateErr


def test_fsm_loop_once_error_recursion(env: StateEnvironment) -> None:
    fsm = FSM(env=env, init_state=MockStateErr, error_state=MockStateErr)

    with pytest.raises(_TestError):
        fsm.loop_once()
    assert fsm.state == MockStateErr
    assert fsm.previous_state == MockStateErr


def test_fsm_change_state(env: StateEnvironment) -> None:
    fsm = FSM(env=env, init_state=MockStateA, error_state=MockStateC)

    # Redundant calls should not break the state machine.
    fsm.change_state(MockStateB)
    fsm.change_state(MockStateB)
    assert fsm.state == MockStateB
    assert fsm.previous_state == MockStateA

    with pytest.raises(InvalidStateError):
        fsm.change_state(MockStateA)


def test_fsm_request_state(env: StateEnvironment) -> None:
    fsm = FSM(env=env, init_state=MockStateA, error_state=MockStateC)

    # Redundant calls should not break the state machine.
    fsm.request_state(MockStateB)
    fsm.request_state(MockStateB)
    fsm.loop_once()
    assert fsm.state == MockStateB
    assert fsm.previous_state == MockStateA

    # Should not raise if we couldn't request the state.
    fsm.request_state(MockStateA)
    fsm.loop_once()
    assert fsm.state == MockStateB
    assert fsm.previous_state == MockStateA

    # Requesting a valid state then changing to an invalid should not explode.
    fsm.request_state(MockStateErr)
    fsm.change_state(MockStateC)
    fsm.loop_once()
    assert fsm.state == MockStateC
    assert fsm.previous_state == MockStateB


def test_fsm_loop_error_explosion(env: StateEnvironment) -> None:
    fsm = FSM(env=env, init_state=MockStateErr, error_state=MockStateA)

    # We can't reach the error state from the broken initial state.
    with pytest.raises(InvalidStateError) as excinfo:
        fsm.loop_once()
    assert isinstance(excinfo.value.__cause__, _TestError)
    assert fsm.state == MockStateErr
    assert fsm.previous_state == MockStateErr


def test_fsm_loop_idle(env: StateEnvironment) -> None:
    fsm = FSM(env=env, init_state=MockStateC, error_state=MockStateC)

    start = threading.Event()
    def wakeup() -> None:
        start.wait()
        time.sleep(0.1)
        with fsm.cond:
            fsm.cond.notify_all()

    thread = threading.Thread(target=wakeup, daemon=True)
    thread.start()

    start.set()
    ret = fsm.loop_until_state([MockStateA])
    assert ret is False



from abc import ABCMeta, abstractmethod
import sys
from typing import ClassVar, NamedTuple, Sequence

from ..handles import PublishHandle
from ..mqtt_spec import MQTTQoS
from ..packet import (
    MQTTPublishPacket,
    MQTTPubAckPacket,
    MQTTPubRecPacket,
    MQTTPubRelPacket,
    MQTTPubCompPacket,
)
from ..property import MQTTPublishProps
from ..topic_alias import AliasPolicy

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class LostMessageError(Exception):
    """Raised when a message is lost from the persistence store and can not be acknowledged."""


class RenderedPacket(NamedTuple):
    """Represents a rendered packet."""
    packet: MQTTPublishPacket | MQTTPubRelPacket
    alias_policy: AliasPolicy


class Persistence(metaclass=ABCMeta):
    """Abstract base class for message persistence."""
    __slots__: ClassVar[Sequence[str]] = tuple()

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type: type | None, exc_value: BaseException | None, traceback: object | None) -> None:
        self.close()

    @abstractmethod
    def __len__(self) -> int:
        """Return the number of outgoing messages in the persistence store."""

    @abstractmethod
    def add(
        self,
        topic: str,
        payload: bytes,
        qos: MQTTQoS,
        retain: bool,
        properties: MQTTPublishProps,
        alias_policy: AliasPolicy,
    ) -> PublishHandle:
        """Add a PUBLISH message to the persistence store."""

    @abstractmethod
    def get(self, count: int) -> Sequence[int]:
        """Get the packet ids of some pending messages from the store."""

    @abstractmethod
    def ack(self, packet: MQTTPubAckPacket | MQTTPubRecPacket | MQTTPubCompPacket) -> None:
        """Ack a PUBLISH or PUBREL message in the persistence store.

        Raises ValueError if the packet_id is not found in the store."""

    @abstractmethod
    def check_rec(self, packet: MQTTPublishPacket) -> bool:
        """Validate that a QoS 2 PUBLISH packet has not already been received.

        Returns True if the packet has not already been received, otherwise False.

        Raises ValueError if the packet is not a QoS 2 PUBLISH packet."""

    @abstractmethod
    def set_rec(self, packet: MQTTPublishPacket) -> None:
        """Indicate that a QoS 2 PUBLISH message has been received.

        Raises ValueError if the packet is not a QoS 2 PUBLISH packet."""

    @abstractmethod
    def rel(self, packet: MQTTPubRelPacket) -> None:
        """Release a QoS 2 PUBLISH message."""

    @abstractmethod
    def render(self, packet_id: int) -> RenderedPacket:
        """Render a PUBLISH message from the persistence store.

        This also indicates to the persistence store that the message is inflight.

        Raises KeyError if the ID is not retained."""

    @abstractmethod
    def clear(self) -> None:
        """Clear the persistence store, discarding all pending messages."""

    @abstractmethod
    def open(self, client_id: str, clear: bool = False) -> None:
        """Indicate to the persistence store that the broker has acknowledged our connection.

        This may clear the persistence store if the client_id is different from the persisted,
        or if clear is True."""

    @abstractmethod
    def close(self) -> None:
        """Finalize and close the persistence store.

        The store must not be used after this call."""



from __future__ import annotations

from typing import Final, Mapping

from ..error import MQTTError
from ..mqtt_spec import MQTTPacketType, MQTTReasonCode
from ..serialization import decode_varint

from .base import MQTTPacket as MQTTPacket
from .connect import MQTTConnectPacket as MQTTConnectPacket
from .connect import MQTTConnAckPacket as MQTTConnAckPacket
from .connect import MQTTDisconnectPacket as MQTTDisconnectPacket
from .publish import MQTTPublishPacket as MQTTPublishPacket
from .publish import MQTTPubAckPacket as MQTTPubAckPacket
from .publish import MQTTPubRecPacket as MQTTPubRecPacket
from .publish import MQTTPubRelPacket as MQTTPubRelPacket
from .publish import MQTTPubCompPacket as MQTTPubCompPacket
from .subscribe import MQTTSubscribePacket as MQTTSubscribePacket
from .subscribe import MQTTSubAckPacket as MQTTSubAckPacket
from .subscribe import MQTTUnsubscribePacket as MQTTUnsubscribePacket
from .subscribe import MQTTUnsubAckPacket as MQTTUnsubAckPacket
from .ping import MQTTPingReqPacket as MQTTPingReqPacket
from .ping import MQTTPingRespPacket as MQTTPingRespPacket
from .ping import PING as PING
from .ping import PONG as PONG
from .auth import MQTTAuthPacket as MQTTAuthPacket


# Map of packet types to their respective classes.
_ControlPacketClasses: Final[Mapping[int, type[MQTTPacket]]] = {
    MQTTPacketType.CONNECT: MQTTConnectPacket,
    MQTTPacketType.CONNACK: MQTTConnAckPacket,
    MQTTPacketType.PUBLISH: MQTTPublishPacket,
    MQTTPacketType.PUBACK: MQTTPubAckPacket,
    MQTTPacketType.PUBREC: MQTTPubRecPacket,
    MQTTPacketType.PUBREL: MQTTPubRelPacket,
    MQTTPacketType.PUBCOMP: MQTTPubCompPacket,
    MQTTPacketType.SUBSCRIBE: MQTTSubscribePacket,
    MQTTPacketType.SUBACK: MQTTSubAckPacket,
    MQTTPacketType.UNSUBSCRIBE: MQTTUnsubscribePacket,
    MQTTPacketType.UNSUBACK: MQTTUnsubAckPacket,
    MQTTPacketType.PINGREQ: MQTTPingReqPacket,
    MQTTPacketType.PINGRESP: MQTTPingRespPacket,
    MQTTPacketType.DISCONNECT: MQTTDisconnectPacket,
    MQTTPacketType.AUTH: MQTTAuthPacket,
}


def decode_packet(data: bytes) -> MQTTPacket:
    """Decode a packet from binary data.

    The packet must be complete and correctly framed."""
    try:
        cls_id = data[0] // 0x10
        decoder = _ControlPacketClasses[cls_id]
    except KeyError:
        raise MQTTError(f"Invalid packet type {cls_id}", MQTTReasonCode.MalformedPacket)
    flags = data[0] % 0x10

    view = memoryview(data)
    length, sz = decode_varint(view[1:])
    offset = sz + 1
    remainder = view[offset:]
    if len(remainder) != length:
        raise MQTTError(f"Invalid length, expected {length} bytes but got {len(remainder)}", MQTTReasonCode.MalformedPacket)
    return decoder.decode(flags, remainder)


def decode_packet_from_parts(head: int, data: memoryview) -> MQTTPacket:
    """Finish decoding a packet which has already been split into parts by an incremental reader."""
    try:
        cls_id = head // 0x10
        decoder = _ControlPacketClasses[cls_id]
    except KeyError:
        raise MQTTError(f"Invalid packet type {cls_id}", MQTTReasonCode.MalformedPacket)
    flags = head % 0x10

    return decoder.decode(flags, data)



#!/usr/bin/env python3
"""This example demonstrates waiting for a published message to be fully acknowledged by the broker
before proceeding to send the next message.

It also demonstrates the debug logging output of the client.
"""

import argparse
import logging

from ohmqtt import Client, MQTTQoS


def main(args: argparse.Namespace) -> None:
    logging.basicConfig(level=logging.DEBUG)
    with Client() as client:

        client.connect(args.address)
        client.wait_for_connect(timeout=5.0)

        for n in range(1, 9):
            client.publish(
                "ohmqtt/examples/publish_wait_for_ack", b"test_payload: " + str(n).encode(), qos=MQTTQoS.Q2
            ).wait_for_ack(timeout=5.0)

        client.disconnect()
        client.wait_for_disconnect(timeout=5.0)


if __name__ == "__main__":
    from .args import parser
    args = parser.parse_args()
    main(args)



import argparse
import time

from .common import add_broker_arguments, get_client, Command, SubParsersT
from .. import MQTTPublishProps


class PublishCommand(Command):
    """Publish a message to a topic."""

    @classmethod
    def register(cls, sub_parsers: SubParsersT) -> None:
        parser = sub_parsers.add_parser(
            "publish", help="Publish a message to a topic"
        )
        add_broker_arguments(parser)
        parser.add_argument("topic", type=str, help="The topic to publish to")
        parser.add_argument("message", type=str, nargs="?", default="", help="The message content to publish")
        parser.add_argument(
            "--qos", type=int, choices=[0, 1, 2], default=0, help="The QoS level for the message"
        )
        parser.add_argument(
            "--retain", action="store_true", help="Set the retain flag for the message"
        )
        parser.add_argument(
            "--payload_format_indicator", action="store_true", help="Indicate that the payload is UTF-8 encoded"
        )
        parser.add_argument(
            "--message_expiry_interval", type=int, help="Set the message expiry interval in seconds"
        )
        parser.add_argument(
            "--content_type", type=str, help="Set the content type of the message"
        )
        parser.add_argument(
            "--response_topic", type=str, help="Set the response topic for the message"
        )
        parser.add_argument(
            "--correlation_data", type=str, help="Set the correlation data for the message (will be encoded as UTF-8)"
        )
        parser.set_defaults(func=cls.execute)

    @classmethod
    def execute(cls, args: argparse.Namespace) -> None:
        topic = args.topic
        message = args.message.encode("utf-8")
        qos = args.qos
        retain = args.retain
        props = MQTTPublishProps()
        props_map = {
            "payload_format_indicator": "PayloadFormatIndicator",
            "message_expiry_interval": "MessageExpiryInterval",
            "content_type": "ContentType",
            "response_topic": "ResponseTopic",
            "correlation_data": "CorrelationData",
        }
        for arg_name, prop_name in props_map.items():
            value = getattr(args, arg_name, None)
            if value is not None:
                setattr(props, prop_name, value)

        if props.CorrelationData is not None:
            props.CorrelationData = props.CorrelationData.encode("utf-8")  # type: ignore

        client = get_client(args)

        handle = client.publish(
            topic,
            message,
            qos=qos,
            retain=retain,
            properties=props,
        )
        client.loop_once(max_wait=0)
        t0 = time.monotonic()
        while qos > 0 and (handle.ack is None and time.monotonic() - t0 < 10.0):
            client.loop_once(max_wait=0.1)  # pragma: no cover
        print(f"Published message to {topic}")
        client.shutdown()
        client.loop_forever()



"""AUTH packets."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Final

from .base import MQTTPacket
from ..error import MQTTError
from ..mqtt_spec import MQTTPacketType, MQTTReasonCode
from ..property import (
    MQTTAuthProps,
)
from ..serialization import (
    encode_varint,
    decode_uint8,
)

HEAD_AUTH: Final = MQTTPacketType.AUTH << 4


@dataclass(match_args=True, slots=True)
class MQTTAuthPacket(MQTTPacket):
    packet_type = MQTTPacketType.AUTH
    props_type = MQTTAuthProps
    reason_code: MQTTReasonCode = MQTTReasonCode.Success
    properties: MQTTAuthProps = field(default_factory=MQTTAuthProps)

    def __str__(self) -> str:
        attrs = [
            f"reason_code={hex(self.reason_code)}",
            f"properties={self.properties}",
        ]
        return f"AUTH[{', '.join(attrs)}]"

    def encode(self) -> bytes:
        # If the reason code is success and there are no properties, the packet can be empty.
        if self.reason_code == MQTTReasonCode.Success and len(self.properties) == 0:
            return HEAD_AUTH.to_bytes(1, "big") + b"\x00"
        encoded = bytearray()
        encoded.append(HEAD_AUTH)
        props = self.properties.encode()
        encoded.extend(encode_varint(len(props) + 1))
        encoded.append(self.reason_code.value)
        encoded.extend(props)
        return bytes(encoded)

    @classmethod
    def decode(cls, flags: int, data: memoryview) -> MQTTAuthPacket:
        if flags != 0:
            raise MQTTError(f"Invalid flags, expected 0 but got {flags}", MQTTReasonCode.MalformedPacket)
        if len(data) == 0:
            # An empty packet means success with no properties.
            return MQTTAuthPacket()
        reason_code, sz = decode_uint8(data)
        props, _ = MQTTAuthProps.decode(data[sz:])
        return MQTTAuthPacket(MQTTReasonCode(reason_code), properties=props)



"""RFC 6455 Websocket protocol library."""

import base64
from enum import IntEnum
import hashlib
import secrets
import sys
from typing import Final

from ..logger import get_logger


logger: Final = get_logger("ohmqtt.connection.wslib")
GUID: Final = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"


class WebsocketError(Exception):
    """Exception raised for WebSocket protocol errors."""


class OpCode(IntEnum):
    """Websocket OpCode values."""
    CONT = 0x00
    TEXT = 0x01
    BINARY = 0x02
    CLOSE = 0x08
    PING = 0x09
    PONG = 0x0A


def generate_nonce() -> str:
    """Generate a random base64-encoded nonce for Websocket handshake."""
    random_bytes = secrets.token_bytes(16)
    return base64.b64encode(random_bytes).decode("utf-8")


def generate_handshake_key(nonce: str) -> str:
    """Generate the expected Sec-WebSocket-Accept key from the server."""
    expected_key = nonce + GUID
    expected_key_digest = hashlib.sha1(expected_key.encode("utf-8")).digest()
    return base64.b64encode(expected_key_digest).decode("utf-8")


def generate_mask() -> bytes:
    """Generate a random 4-byte mask for Websocket frames."""
    # Per RFC 6455, Section 5.3, the mask must come from a strong source of entropy.
    return secrets.token_bytes(4)


def apply_mask(mask: bytes, data: bytes) -> bytes:
    """Apply a WebSocket mask to the input data."""
    # This is a performance-critical method.
    # Do not loop over the payload data.
    # In fact, just use the pure Python implementation from the websockets library,
    #   which is based on Will McGugan's implementation.
    # See: https://github.com/python-websockets/websockets/commit/c7fc0d36bd8ea2aeb7c4321f53d208fb1297db85
    assert len(mask) == 4, "Mask must be 4 bytes"
    data_int = int.from_bytes(data, sys.byteorder)
    mask_repeated = mask * (len(data) // 4) + mask[:len(data) % 4]
    mask_int = int.from_bytes(mask_repeated, sys.byteorder)
    return (data_int ^ mask_int).to_bytes(len(data), sys.byteorder)


def frame_ws_data(opcode: OpCode, payload: bytes, do_mask: bool = True) -> bytes:
    """Frame data for WebSocket transport.

    All data frames sent to the broker will be singular, final frames."""
    header = bytes([0x80 | opcode.value])
    length = len(payload)
    mask_bit = 0x80 if do_mask else 0x00
    if length <= 125:
        length_encoded = bytes([length | mask_bit])
    elif length <= 0xffff:
        length_encoded = bytes([126 | mask_bit]) + length.to_bytes(2, "big")
    else:
        length_encoded = bytes([127 | mask_bit]) + length.to_bytes(8, "big")
    if do_mask:
        mask = generate_mask()
        payload = apply_mask(mask, payload)
        return header + length_encoded + mask + payload
    return header + length_encoded + payload


def deframe_ws_data(buffer: bytearray | bytes) -> tuple[OpCode, bytes, bool, int] | None:
    """Deframe data from WebSocket transport.

    Returns a tuple of (OpCode, payload, masked) if a complete frame is available,
        or None if more data is needed."""
    if len(buffer) < 2:
        logger.debug("Deframe needs at least 2 bytes, has %d", len(buffer))
        return None

    fin = (buffer[0] & 0x80) != 0
    if not fin:
        raise WebsocketError("Fragmented WebSocket frames are not supported")
    opcode = OpCode(buffer[0] & 0x0F)
    masked = (buffer[1] & 0x80) != 0
    payload_length = buffer[1] & 0x7F

    index = 2
    if payload_length == 126:
        if len(buffer) < index + 2:
            logger.debug("Deframe needs at least %d bytes for extended payload length, has %d", index + 2, len(buffer))
            return None
        payload_length = int.from_bytes(buffer[index:index + 2], "big")
        index += 2
    elif payload_length == 127:
        if len(buffer) < index + 8:
            logger.debug("Deframe needs at least %d bytes for extended payload length, has %d", index + 8, len(buffer))
            return None
        payload_length = int.from_bytes(buffer[index:index + 8], "big")
        index += 8

    frame_length = index + masked * 4 + payload_length
    if len(buffer) < frame_length:
        logger.debug("Deframe needs at least %d bytes for complete frame, has %d", frame_length, len(buffer))
        return None

    mask: bytes | None = None
    if masked:
        mask = bytes(buffer[index:index + 4])
        index += 4

    payload = bytes(buffer[index:frame_length])

    if masked and mask is not None:
        payload = apply_mask(mask, payload)

    return opcode, payload, masked, frame_length



import time


_time = time.monotonic

class Timeout:
    """A simple timer class for getting timeouts since the last event."""
    __slots__ = ("_mark", "interval")

    def __init__(self, interval: float | None = None) -> None:
        self.interval = interval
        self.mark()

    def mark(self) -> None:
        """Mark an event."""
        self._mark = _time()

    def get_timeout(self, max_wait: float | None = None) -> float | None:
        """Get the difference between the interval and the last mark.

        If the value would be negative, returns 0.

        If the interval is None, returns None.

        If max_wait is not None, returns the minimum of max_wait and the interval."""
        now = _time()
        if max_wait is not None:
            if self.interval is None:
                return max_wait
            return min(max_wait, self.interval - (now - self._mark))
        if self.interval is None:
            return None
        return max(0, self.interval - (now - self._mark))

    def exceeded(self) -> bool:
        """Check if the timeout has been exceeded.

        If the interval is None, always returns False."""
        if self.interval is None:
            return False
        return self.interval - (_time() - self._mark) <= 0



from __future__ import annotations

import socket
from typing import Final, TYPE_CHECKING

from .base import FSMState
from .closed import ClosedState
from .mqtt_handshake_connect import MQTTHandshakeConnectState
from .tls_handshake import TLSHandshakeState
from .websocket_handshake_request import WebsocketHandshakeRequestState
from ..types import ConnectParams, StateData, StateEnvironment
from ...logger import get_logger
from ...platform import AF_UNIX, HAS_AF_UNIX

if TYPE_CHECKING:
    from ..fsm import FSM

logger: Final = get_logger("ohmqtt.connection.states.connecting")


def _get_socket(family: socket.AddressFamily) -> socket.socket:
    """Get a socket object.

    This is patched in tests to use a mock or loopback socket."""
    return socket.socket(family, socket.SOCK_STREAM)


class ConnectingState(FSMState):
    """Connecting to the broker."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        state_data.keepalive.keepalive_interval = params.keepalive_interval
        state_data.timeout.interval = params.connect_timeout
        state_data.timeout.mark()
        state_data.connack = None
        state_data.disconnect_rc = None
        state_data.sock = _get_socket(params.address.family)
        if params.address.family in (socket.AF_INET, socket.AF_INET6):
            state_data.sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, params.tcp_nodelay)
        state_data.decoder.reset()
        state_data.ws_decoder.reset()
        state_data.ws_handshake_buffer.clear()
        state_data.ws_nonce = ""
        state_data.open_called = False
        with fsm.selector:
            fsm.selector.change_sock(state_data.sock)

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:

        if state_data.timeout.exceeded():
            logger.debug("Connect timeout")
            fsm.change_state(ClosedState)
            return True

        try:
            address = params.address
            if HAS_AF_UNIX and address.family == AF_UNIX:
                state_data.sock.connect(address.host)
            else:
                state_data.sock.connect((address.host, address.port))
        except ConnectionError as exc:
            logger.error("Failed to connect to broker: %s", exc)
            fsm.change_state(ClosedState)
            return True
        except (BlockingIOError, OSError):
            pass  # Either in progress or already connected, select will reveal which.

        with fsm.selector:
            timeout = state_data.timeout.get_timeout(max_wait)
            _, writable = fsm.selector.select(write=True, timeout=timeout)
        if writable:
            logger.debug("Socket connected to broker")
            if params.address.use_tls:
                fsm.change_state(TLSHandshakeState)
            elif params.address.is_websocket():
                fsm.change_state(WebsocketHandshakeRequestState)
            else:
                fsm.change_state(MQTTHandshakeConnectState)
            state_data.sock.setblocking(False)
            return True
        return False



from __future__ import annotations

from http.client import HTTPResponse
from io import BytesIO
import socket
import ssl
from typing import cast, Final, TYPE_CHECKING

from .base import FSMState
from .closed import ClosedState
from .mqtt_handshake_connect import MQTTHandshakeConnectState
from ..types import ConnectParams, StateData, StateEnvironment
from ..wslib import generate_handshake_key
from ...logger import get_logger

if TYPE_CHECKING:
    from ..fsm import FSM

logger: Final = get_logger("ohmqtt.connection.states.mqtt_handshake_connack")


class MockSocket:
    """Wrapper class to provide a makefile method for a buffer.

    This lets us use HTTPResponse on a BytesIO buffer."""
    def __init__(self, buffer: BytesIO) -> None:
        self.buffer = buffer

    def makefile(self, mode: str = "r") -> socket.SocketIO:
        return cast(socket.SocketIO, self.buffer)


class WebsocketHandshakeResponseState(FSMState):
    """Receiving Websocket handshake response from the broker."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        state_data.ws_handshake_buffer.clear()

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        if state_data.timeout.exceeded():
            logger.debug("Websocket handshake response keepalive timeout")
            fsm.change_state(ClosedState)
            return True

        try:
            state_data.ws_handshake_buffer.extend(state_data.sock.recv(0xffff))
        except (BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError):
            logger.debug("Websocket handshake response recv would block, waiting for readable")

        validation_result = cls.validate_response(state_data.ws_nonce, state_data.ws_handshake_buffer)
        if validation_result is True:
            state_data.ws_handshake_buffer.clear()
            fsm.change_state(MQTTHandshakeConnectState)
            return True
        if validation_result is False:
            state_data.ws_handshake_buffer.clear()
            fsm.change_state(ClosedState)
            return True

        # Incomplete response, wait for more data.
        with fsm.selector:
            timeout = state_data.timeout.get_timeout(max_wait)
            fsm.selector.select(read=True, timeout=timeout)
        return False

    @classmethod
    def validate_response(cls, nonce: str, buffer: bytearray) -> bool | None:
        """Validate the websocket handshake response.

        Return True if valid, False if invalid, None if incomplete."""
        try:
            bio = BytesIO(buffer)
            sio = cast(socket.socket, MockSocket(bio))
            response = HTTPResponse(sio)
            response.begin()
            if response.status != 101:
                logger.error("Websocket handshake failed with status %d", response.status)
                return False
            accept_key = response.getheader("Sec-WebSocket-Accept")
            expected_key = generate_handshake_key(nonce)
            if accept_key is None or accept_key != expected_key:
                logger.error("Websocket handshake failed: invalid Sec-WebSocket-Accept")
                return False
            protocol = response.getheader("Sec-WebSocket-Protocol")
            if protocol != "mqtt":
                logger.error("Websocket handshake failed: invalid Sec-WebSocket-Protocol")
                return False
            return True
        except Exception:
            logger.debug("Incomplete websocket handshake response, waiting for more data")
            return None



from __future__ import annotations

from typing import ClassVar, Sequence, TYPE_CHECKING

from ..types import ConnectParams, StateData, StateEnvironment

if TYPE_CHECKING:
    from ..fsm import FSM


class FSMState:
    """A finite state in the FSM."""
    request_from: ClassVar[Sequence[type[FSMState]]] = ()
    transitions_to: ClassVar[Sequence[type[FSMState]]] = ()

    def __init__(self) -> None:
        raise TypeError("Do not instantiate FSMStates")

    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        """Called when entering the state.

        This method must not block."""

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        """Called when handling the state.

        This method may block if max_wait is >0 or None.

        :return: True if the state is finished."""
        return True

    @classmethod
    def can_request_from(cls, prev_state: type[FSMState]) -> bool:
        """Check if the FSM can request this state from the given previous state.

        :param prev_state: The previous state.
        :return: True if the FSM can request this state from the given previous state, False otherwise."""
        return prev_state in cls.request_from

    @classmethod
    def can_transition_to(cls, next_state: type[FSMState]) -> bool:
        """Check if the FSM can transition to the given state.

        :param next_state: The state to check.
        :return: True if the FSM can transition to the given state, False otherwise."""
        return next_state in cls.transitions_to


    @classmethod
    def is_final_state(cls) -> bool:
        """Check if this state is a final state.

        :return: True if this state is a final state, False otherwise."""
        return not cls.transitions_to



#!/usr/bin/env python3
"""This example demonstrates publishing a retained message,
    then subscribing to the topic and receiving the retained message."""

import argparse
from queue import Queue

from ohmqtt import Client, MQTTPublishPacket, MQTTQoS


def main(args: argparse.Namespace) -> None:
    with Client() as client:

        client.connect("localhost")
        client.wait_for_connect(timeout=5.0)
        print("*** Connected to broker")

        pub = client.publish("ohmqtt/examples/publish_retain", b"test_payload", qos=MQTTQoS.Q1, retain=True)
        pub.wait_for_ack(timeout=5.0)

        q: Queue[MQTTPublishPacket] = Queue()
        def callback(_: Client, msg: MQTTPublishPacket) -> None:
            q.put(msg)
        client.subscribe("ohmqtt/examples/publish_retain", callback)
        msg = q.get(timeout=5.0)
        assert msg.topic == "ohmqtt/examples/publish_retain"
        assert msg.payload == b"test_payload"
        assert msg.qos == 1
        assert msg.retain
        print(f"*** Received retained message: {msg!s}")

        client.disconnect()
        client.wait_for_disconnect(timeout=5.0)
        print("*** Disconnected from broker")


if __name__ == "__main__":
    from .args import parser
    args = parser.parse_args()
    main(args)



import threading
from unittest.mock import Mock, MagicMock

import pytest
from pytest_mock import MockerFixture

from ohmqtt.client import Client
from ohmqtt.connection import Connection, MessageHandlers
from ohmqtt.mqtt_spec import MQTTReasonCode, MQTTQoS
from ohmqtt.packet import MQTTAuthPacket
from ohmqtt.property import MQTTAuthProps, MQTTPublishProps
from ohmqtt.session import Session
from ohmqtt.subscriptions import Subscriptions, SubscribeCallback
from ohmqtt.topic_alias import AliasPolicy


@pytest.fixture
def mock_connection(mocker: MockerFixture) -> Mock:
    """Mock the Connection class."""
    mock_connection = mocker.MagicMock(spec=Connection)
    mock_connection.fsm.lock.__enter__.side_effect = lambda: None
    mocker.patch("ohmqtt.client.Connection", return_value=mock_connection)
    return mock_connection  # type: ignore[no-any-return]


@pytest.fixture
def mock_handlers(mocker: MockerFixture) -> MagicMock:
    """Mock the MessageHandlers class."""
    mock_handlers = mocker.MagicMock(spec=MessageHandlers)
    mock_handlers.__enter__.return_value = mock_handlers
    mocker.patch("ohmqtt.connection.MessageHandlers", return_value=mock_handlers)
    return mock_handlers  # type: ignore[no-any-return]


@pytest.fixture
def mock_session(mocker: MockerFixture, mock_subscriptions: Mock) -> Mock:
    """Mock the Session class."""
    mock_session = mocker.Mock(spec=Session)
    mock_session.subscriptions = mock_subscriptions
    mocker.patch("ohmqtt.client.Session", return_value=mock_session)
    return mock_session  # type: ignore[no-any-return]


@pytest.fixture
def mock_subscriptions(mocker: MockerFixture) -> Mock:
    """Mock the Subscriptions class."""
    mock_subscriptions = mocker.Mock(spec=Subscriptions)
    mocker.patch("ohmqtt.client.Subscriptions", return_value=mock_subscriptions)
    return mock_subscriptions  # type: ignore[no-any-return]


@pytest.fixture
def mock_thread(mocker: MockerFixture) -> Mock:
    """Mock the threading.Thread class."""
    mock_thread = mocker.Mock(threading.Thread)
    mocker.patch("threading.Thread", return_value=mock_thread)
    return mock_thread  # type: ignore[no-any-return]


def test_client_connect(mocker: MockerFixture, mock_connection: Mock, mock_handlers: MagicMock,
                       mock_session: Mock, mock_subscriptions: Mock) -> None:
    client = Client()
    client.connect("localhost")
    mock_connection.connect.assert_called_once()


def test_client_disconnect(mocker: MockerFixture, mock_connection: Mock, mock_handlers: MagicMock,
                          mock_session: Mock, mock_subscriptions: Mock) -> None:
    client = Client()
    client.disconnect()
    mock_connection.disconnect.assert_called_once()


def test_client_shutdown(mocker: MockerFixture, mock_connection: Mock, mock_handlers: MagicMock,
                        mock_session: Mock, mock_subscriptions: Mock) -> None:
    client = Client()
    client.shutdown()
    mock_connection.shutdown.assert_called_once()


@pytest.mark.parametrize("payload", [b"test_payload", bytearray(b"test_payload"), "test_payload"])
@pytest.mark.parametrize("qos", [0, 1, 2, MQTTQoS.Q0, MQTTQoS.Q1, MQTTQoS.Q2])
def test_client_publish(payload: bytes | bytearray | str, qos: int | MQTTQoS, mocker: MockerFixture,
                        mock_connection: Mock, mock_handlers: MagicMock, mock_session: Mock, mock_subscriptions: Mock) -> None:
    client = Client()

    mock_session.publish.return_value = mocker.Mock()
    publish_handle = client.publish(
        "test/topic",
        payload,
        qos=qos,
        retain=True,
        properties=MQTTPublishProps(
            MessageExpiryInterval=60,
            ResponseTopic="response/topic",
            CorrelationData=b"correlation_data",
            UserProperty=[("key", "value")],
        ),
        alias_policy=AliasPolicy.ALWAYS,
    )
    assert publish_handle == mock_session.publish.return_value
    expected_qos = MQTTQoS(qos) if not isinstance(qos, MQTTQoS) else qos
    expected_payload = payload.encode("utf-8") if isinstance(payload, str) else bytes(payload)
    mock_session.publish.assert_called_once_with(
        "test/topic",
        expected_payload,
        qos=expected_qos,
        retain=True,
        properties=MQTTPublishProps(
            MessageExpiryInterval=60,
            ResponseTopic="response/topic",
            CorrelationData=b"correlation_data",
            UserProperty=[("key", "value")],
        ),
        alias_policy=AliasPolicy.ALWAYS,
    )


@pytest.mark.parametrize("max_qos", [0, 1, 2, MQTTQoS.Q0, MQTTQoS.Q1, MQTTQoS.Q2])
def test_client_subscribe(max_qos: int | MQTTQoS, mocker: MockerFixture, mock_connection: Mock, mock_handlers: MagicMock,
                         mock_session: Mock, mock_subscriptions: Mock) -> None:
    client = Client()

    callback: SubscribeCallback = lambda _client, _packet: None
    mock_subscriptions.subscribe.return_value = mocker.Mock()
    sub_handle = client.subscribe("test/+", callback, max_qos=max_qos)
    assert sub_handle == mock_subscriptions.subscribe.return_value
    mock_subscriptions.subscribe.assert_called_once()
    expected_max_qos = MQTTQoS(max_qos) if not isinstance(max_qos, MQTTQoS) else max_qos
    assert mock_subscriptions.subscribe.call_args[0][0] == "test/+"
    assert mock_subscriptions.subscribe.call_args[0][1] == callback
    assert mock_subscriptions.subscribe.call_args_list[0].kwargs["max_qos"] == expected_max_qos


def test_client_unsubscribe(mocker: MockerFixture, mock_connection: Mock, mock_handlers: MagicMock,
                            mock_session: Mock, mock_subscriptions: Mock) -> None:
    client = Client()

    mock_subscriptions.unsubscribe.return_value = mocker.Mock()
    unsub_handle = client.unsubscribe("test/topic")
    assert unsub_handle == mock_subscriptions.unsubscribe.return_value
    mock_subscriptions.unsubscribe.assert_called_once()
    assert mock_subscriptions.unsubscribe.call_args[0][0] == "test/topic"


@pytest.mark.parametrize("auth_method", ["test_method", None])
@pytest.mark.parametrize("auth_data", [b"test_data", None])
@pytest.mark.parametrize("reason_string", ["reason", None])
@pytest.mark.parametrize("user_properties", [[("key", "value")], None])
def test_client_auth(auth_method: str | None, auth_data: bytes | None,
                     reason_string: str | None, user_properties: list[tuple[str, str]] | None,
                     mocker: MockerFixture, mock_connection: Mock, mock_handlers: MagicMock,
                     mock_session: Mock, mock_subscriptions: Mock) -> None:
    client = Client()

    client.auth(
        reason_code=MQTTReasonCode.ReAuthenticate,
        authentication_method=auth_method,
        authentication_data=auth_data,
        reason_string=reason_string,
        user_properties=user_properties,
    )
    expected_props = MQTTAuthProps()
    if auth_method is not None:
        expected_props.AuthenticationMethod = auth_method
    if auth_data is not None:
        expected_props.AuthenticationData = auth_data
    if reason_string is not None:
        expected_props.ReasonString = reason_string
    if user_properties is not None:
        expected_props.UserProperty = user_properties
    mock_connection.send.assert_called_once_with(MQTTAuthPacket(
        reason_code=MQTTReasonCode.ReAuthenticate,
        properties=expected_props,
    ))


def test_client_wait_for_connect(mocker: MockerFixture, mock_connection: Mock, mock_handlers: MagicMock,
                                mock_session: Mock, mock_subscriptions: Mock) -> None:
    client = Client()

    mock_connection.wait_for_connect.return_value = True
    client.wait_for_connect(0.1)
    mock_connection.wait_for_connect.assert_called_once_with(0.1)

    mock_connection.wait_for_connect.return_value = False
    with pytest.raises(TimeoutError):
        client.wait_for_connect(0.1)


def test_client_wait_for_disconnect(mocker: MockerFixture, mock_connection: Mock, mock_handlers: MagicMock,
                                   mock_session: Mock, mock_subscriptions: Mock) -> None:
    client = Client()

    mock_connection.wait_for_disconnect.return_value = True
    client.wait_for_disconnect(0.1)
    mock_connection.wait_for_disconnect.assert_called_once_with(0.1)

    mock_connection.wait_for_disconnect.return_value = False
    with pytest.raises(TimeoutError):
        client.wait_for_disconnect(0.1)


def test_client_wait_for_shutdown(mocker: MockerFixture, mock_connection: Mock, mock_handlers: MagicMock,
                                  mock_session: Mock, mock_subscriptions: Mock) -> None:
    client = Client()

    mock_connection.wait_for_shutdown.return_value = True
    client.wait_for_shutdown(0.1)
    mock_connection.wait_for_shutdown.assert_called_once_with(0.1)

    mock_connection.wait_for_shutdown.return_value = False
    with pytest.raises(TimeoutError):
        client.wait_for_shutdown(0.1)


def test_client_start_loop(mocker: MockerFixture, mock_connection: Mock, mock_handlers: MagicMock,
                          mock_session: Mock, mock_subscriptions: Mock, mock_thread: Mock) -> None:
    client = Client()
    client.start_loop()
    mock_thread.start.assert_called_once()
    with pytest.raises(RuntimeError):
        client.start_loop()


def test_client_loop_once(mocker: MockerFixture, mock_connection: Mock, mock_handlers: MagicMock,
                         mock_session: Mock, mock_subscriptions: Mock) -> None:
    client = Client()
    client.loop_once(0.1)
    mock_connection.loop_once.assert_called_once_with(0.1)


def test_client_loop_forever(mocker: MockerFixture, mock_connection: Mock, mock_handlers: MagicMock,
                            mock_session: Mock, mock_subscriptions: Mock) -> None:
    client = Client()
    client.loop_forever()
    mock_connection.loop_forever.assert_called_once_with()


def test_client_loop_until_connected(mocker: MockerFixture, mock_connection: Mock, mock_handlers: MagicMock,
                                    mock_session: Mock, mock_subscriptions: Mock) -> None:
    client = Client()
    client.loop_until_connected(0.1)
    mock_connection.loop_until_connected.assert_called_once_with(0.1)


def test_client_is_connected(mocker: MockerFixture, mock_connection: Mock, mock_handlers: MagicMock,
                            mock_session: Mock, mock_subscriptions: Mock) -> None:
    client = Client()
    mock_connection.is_connected.return_value = True
    assert client.is_connected() is True
    mock_connection.is_connected.assert_called_once()
    mock_connection.is_connected.reset_mock()

    mock_connection.is_connected.return_value = False
    assert client.is_connected() is False
    mock_connection.is_connected.assert_called_with()


def test_client_handle_auth(mocker: MockerFixture, mock_connection: Mock, mock_handlers: MagicMock,
                           mock_session: Mock, mock_subscriptions: Mock) -> None:
    client = Client()

    auth_packet = MQTTAuthPacket()
    client.handle_auth(auth_packet)


def test_client_slots(mocker: MockerFixture, mock_connection: Mock, mock_handlers: MagicMock,
                     mock_session: Mock, mock_subscriptions: Mock) -> None:
    """Test that the client slots are set correctly."""
    with Client() as client:
        assert not hasattr(client, "__dict__")
        assert hasattr(client, "__weakref__")
        assert all(hasattr(client, attr) for attr in client.__slots__)



import logging
from typing import Final

ohmqtt_logger: Final = logging.getLogger("ohmqtt")


def set_log_level(level: int) -> None:
    """Set the log level for the ohmqtt logger."""
    ohmqtt_logger.setLevel(level)


def get_logger(name: str) -> logging.Logger:
    """Get a logger with the specified name."""
    return ohmqtt_logger.getChild(name)



#!/usr/bin/env python3
"""This example demonstrates connecting to a broker, publishing messages to a topic, and then disconnecting."""

import argparse
import time

from ohmqtt import Client, MQTTQoS


def main(args: argparse.Namespace) -> None:
    with Client() as client:

        print("*** Connecting to broker...")
        client.connect(args.address)

        client.wait_for_connect(timeout=5.0)
        print("*** Connected to broker")

        for n in range(1, 9):
            client.publish("ohmqtt/examples/publish", b"test_payload: " + str(n).encode(), qos=MQTTQoS.Q0)
            print(f"*** Published message {n}")
            time.sleep(1.0)

        print("*** Disconnecting from broker...")
        client.disconnect()

        client.wait_for_disconnect(timeout=5.0)
        print("*** Disconnected from broker")


if __name__ == "__main__":
    from .args import parser
    args = parser.parse_args()
    main(args)
